@@ -14,27 +14,26 @@
 #
 # Author: Jonas Borgström <jonas@edgewall.com>
 
-import inspect
 import os
 import pkg_resources
 import re
 import shutil
-import sys
 
 from genshi import HTML
 from genshi.builder import tag
 
-from trac import __version__ as TRAC_VERSION
 from trac.admin.api import IAdminPanelProvider
 from trac.core import *
+from trac.loader import get_plugin_info, get_plugins_dir
 from trac.perm import PermissionSystem, IPermissionRequestor
-from trac.util import get_pkginfo, get_module_path
 from trac.util.compat import partial
-from trac.util.text import to_unicode
-from trac.util.translation import _
+from trac.util.text import exception_to_unicode
+from trac.util.translation import _, ngettext
 from trac.web import HTTPNotFound, IRequestHandler
-from trac.web.chrome import add_script, add_stylesheet, add_warning, Chrome, \
-                            INavigationContributor, ITemplateProvider
+from trac.web.chrome import add_notice, add_stylesheet, \
+                            add_warning, Chrome, INavigationContributor, \
+                            ITemplateProvider
+from trac.wiki.formatter import format_to_html
 
 try:
     from webadmin import IAdminPageProvider
@@ -43,7 +42,7 @@ except ImportError:
 
 
 class AdminModule(Component):
-    """Web administration interface."""
+    """Web administration interface provider and panel manager."""
 
     implements(INavigationContributor, IRequestHandler, ITemplateProvider)
 
@@ -83,7 +82,11 @@ class AdminModule(Component):
             raise HTTPNotFound(_('No administration panels available'))
 
         def _panel_order(p1, p2):
-            if p1[0] == 'general':
+            if p1[::2] == ('general', 'basics'):
+                return -1
+            elif p2[::2] == ('general', 'basics'):
+                return 1
+            elif p1[0] == 'general':
                 if p2[0] == 'general':
                     return cmp(p1[1:], p2[1:])
                 return -1
@@ -157,7 +160,7 @@ class AdminModule(Component):
         providers = {}
 
         for provider in self.panel_providers:
-            p = list(provider.get_admin_panels(req))
+            p = list(provider.get_admin_panels(req) or [])
             for panel in p:
                 providers[(panel[0], panel[2])] = provider
             panels += p
@@ -173,6 +176,23 @@ class AdminModule(Component):
         return panels, providers
 
 
+def _save_config(config, req, log, notices=None):
+    """Try to save the config, and display either a success notice or a
+    failure warning.
+    """
+    try:
+        config.save()
+        if notices is None:
+            notices = [_('Your changes have been saved.')]
+        for notice in notices:
+            add_notice(req, notice)
+    except Exception, e:
+        log.error('Error writing to trac.ini: %s', exception_to_unicode(e))
+        add_warning(req, _('Error writing to trac.ini, make sure it is '
+                           'writable by the web server. Your changes have '
+                           'not been saved.'))
+
+
 class BasicsAdminPanel(Component):
 
     implements(IAdminPanelProvider)
@@ -189,7 +209,7 @@ class BasicsAdminPanel(Component):
         if req.method == 'POST':
             for option in ('name', 'url', 'descr'):
                 self.config.set('project', option, req.args.get(option))
-            self.config.save()
+            _save_config(self.config, req, self.log)
             req.redirect(req.href.admin(cat, page))
 
         data = {
@@ -197,6 +217,7 @@ class BasicsAdminPanel(Component):
             'description': self.env.project_description,
             'url': self.env.project_url
         }
+        Chrome(self.env).add_textarea_grips(req)
         return 'admin_basics.html', {'project': data}
 
 
@@ -217,7 +238,7 @@ class LoggingAdminPanel(Component):
         log_dir = os.path.join(self.env.path, 'log')
 
         log_types = [
-            dict(name='', label=_('None'), selected=False, disabled=False),
+            dict(name='none', label=_('None'), selected=log_type == 'none', disabled=False),
             dict(name='stderr', label=_('Console'),
                  selected=log_type == 'stderr', disabled=False),
             dict(name='file', label=_('File'), selected=log_type == 'file',
@@ -235,38 +256,37 @@ class LoggingAdminPanel(Component):
             changed = False
 
             new_type = req.args.get('log_type')
-            if new_type and new_type not in ('stderr', 'file', 'syslog',
-                                             'eventlog'):
+            if new_type not in [t['name'] for t in log_types]:
                 raise TracError(
                     _('Unknown log type %(type)s', type=new_type),
                     _('Invalid log type')
                 )
             if new_type != log_type:
-                self.config.set('logging', 'log_type', new_type or 'none')
+                self.config.set('logging', 'log_type', new_type)
                 changed = True
                 log_type = new_type
 
-            if log_type:
+            if log_type == 'none':
+                self.config.remove('logging', 'log_level')
+                changed = True
+            else:
                 new_level = req.args.get('log_level')
-                if new_level and new_level not in log_levels:
+                if new_level not in log_levels:
                     raise TracError(
                         _('Unknown log level %(level)s', level=new_level),
                         _('Invalid log level'))
-                if new_level and new_level != log_level:
+                if new_level != log_level:
                     self.config.set('logging', 'log_level', new_level)
                     changed = True
-                    log_evel = new_level
-            else:
-                self.config.remove('logging', 'log_level')
-                changed = True
+                    log_level = new_level
 
             if log_type == 'file':
                 new_file = req.args.get('log_file', 'trac.log')
                 if new_file != log_file:
-                    self.config.set('logging', 'log_file', new_file or '')
+                    self.config.set('logging', 'log_file', new_file)
                     changed = True
                     log_file = new_file
-                if log_type == 'file' and not log_file:
+                if not log_file:
                     raise TracError(_('You must specify a log file'),
                                     _('Missing field'))
             else:
@@ -274,7 +294,7 @@ class LoggingAdminPanel(Component):
                 changed = True
 
             if changed:
-                self.config.save()
+                _save_config(self.config, req, self.log),
             req.redirect(req.href.admin(cat, page))
 
         data = {
@@ -322,12 +342,14 @@ class PermissionAdminPanel(Component):
                 req.perm.require(action)
                 if (subject, action) not in all_permissions:
                     perm.grant_permission(subject, action)
+                    add_notice(req, _('The subject %(subject)s has been '
+                                      'granted the permission %(action)s.',
+                                      subject=subject, action=action))
                     req.redirect(req.href.admin(cat, page))
                 else:
-                    add_warning(req,
-                            _('Permission "%(action)s" was already granted '
-                            'to "%(subject)s"', action=action,
-                            subject=subject))
+                    add_warning(req, _('The permission %(action)s was already '
+                                       'granted to %(subject)s.',
+                                       action=action, subject=subject))
 
             # Add subject to group
             elif req.args.get('add') and subject and group:
@@ -339,13 +361,16 @@ class PermissionAdminPanel(Component):
                             % (subject, group, action))
                     else:
                         req.perm.require(action)
-                if (subject,group) not in all_permissions:
+                if (subject, group) not in all_permissions:
                     perm.grant_permission(subject, group)
+                    add_notice(req, _('The subject %(subject)s has been added '
+                                      'to the group %(group)s.',
+                                      subject=subject, group=group))
                     req.redirect(req.href.admin(cat, page))
                 else:
-                    add_warning(req, 
-                                _('"%(subject)s" was already added to group '
-                                '"%(group)s"', subject=subject, group=group))
+                    add_warning(req, _('The subject %(subject)s was already '
+                                       'added to the group %(group)s.',
+                                       subject=subject, group=group))
 
             # Remove permissions action
             elif req.args.get('remove') and req.args.get('sel'):
@@ -356,6 +381,8 @@ class PermissionAdminPanel(Component):
                     subject, action = key.split(':', 1)
                     if (subject, action) in perm.get_all_permissions():
                         perm.revoke_permission(subject, action)
+                add_notice(req, _('The selected permissions have been '
+                                  'revoked.'))
                 req.redirect(req.href.admin(cat, page))
 
         return 'admin_perms.html', {
@@ -368,14 +395,6 @@ class PluginAdminPanel(Component):
 
     implements(IAdminPanelProvider)
 
-    # Ideally, this wouldn't be hard-coded like this
-    required_components = ('AboutModule', 'DefaultPermissionGroupProvider',
-        'Environment', 'EnvironmentSetup', 'PermissionSystem',
-        'RequestDispatcher', 'Mimeview', 'Chrome')
-
-    def __init__(self):
-        self.trac_path = get_module_path(sys.modules['trac.core'])
-
     # IAdminPanelProvider methods
 
     def get_admin_panels(self, req):
@@ -460,109 +479,63 @@ class PluginAdminPanel(Component):
         """Update component enablement."""
         components = req.args.getlist('component')
         enabled = req.args.getlist('enable')
-        changes = False
+        added, removed = [], []
 
         # FIXME: this needs to be more intelligent and minimize multiple
         # component names to prefix rules
 
         for component in components:
-            is_enabled = self.env.is_component_enabled(component)
-            if is_enabled != (component in enabled):
+            is_enabled = bool(self.env.is_component_enabled(component))
+            must_enable = component in enabled
+            if is_enabled != must_enable:
                 self.config.set('components', component,
                                 is_enabled and 'disabled' or 'enabled')
                 self.log.info('%sabling component %s',
                               is_enabled and 'Dis' or 'En', component)
-                changes = True
+                if must_enable:
+                    added.append(component)
+                else:
+                    removed.append(component)
 
-        if changes:
-            self.config.save()
+        if added or removed:
+            def make_list(items):
+                parts = [item.rsplit('.', 1) for item in items]
+                return tag.table(tag.tbody(
+                    tag.tr(tag.td(c, class_='trac-name'),
+                           tag.td('(%s.*)' % m, class_='trac-name'))
+                    for m, c in parts), class_='trac-pluglist')
+
+            added.sort()
+            removed.sort()
+            notices = []
+            if removed:
+                msg = ngettext('The following component has been disabled:',
+                               'The following components have been disabled:',
+                               len(removed))
+                notices.append(tag(msg, make_list(removed)))
+            if added:
+                msg = ngettext('The following component has been enabled:',
+                               'The following components have been enabled:',
+                               len(added))
+                notices.append(tag(msg, make_list(added)))
+
+            _save_config(self.config, req, self.log, notices)
 
     def _render_view(self, req):
-        plugins = {}
-        plugins_dir = os.path.realpath(os.path.join(self.env.path, 'plugins'))
-        plugins_dir = os.path.normcase(plugins_dir) # needs to match loader.py
+        plugins = get_plugin_info(self.env, include_core=True)
 
-        from trac.core import ComponentMeta
-        for component in ComponentMeta._components:
-            module = sys.modules[component.__module__]
-
-            dist = self._find_distribution(module)
-            plugin_filename = None
-            if os.path.realpath(os.path.dirname(dist.location)) == plugins_dir:
-                plugin_filename = os.path.basename(dist.location)
-
-            description = inspect.getdoc(component)
-            if description:
-                description = to_unicode(description).split('.', 1)[0] + '.'
-
-            if dist.project_name not in plugins:
-                readonly = True
-                if plugin_filename and os.access(dist.location,
-                                                 os.F_OK + os.W_OK):
-                    readonly = False
-                # retrieve plugin metadata
-                info = get_pkginfo(dist)
-                if not info:
-                    info = {'summary': description}
-                    for k in ('author author_email home_page url license trac'
-                              .split()):
-                        v = getattr(module, k, '')
-                        if v:
-                            if k == 'home_page' or k == 'url':
-                                k = 'home_page'
-                                v = v.replace('$', '').replace('URL: ', '') 
-                            info[k] = v
-                # retrieve plugin version info
-                version = dist.version
-                if not version:
-                    version = (getattr(module, 'version', '') or
-                               getattr(module, 'revision', ''))
-                    # special handling for "$Rev$" strings
-                    version = version.replace('$', '').replace('Rev: ', 'r') 
-                plugins[dist.project_name] = {
-                    'name': dist.project_name, 'version': version,
-                    'path': dist.location, 'description': description,
-                    'plugin_filename': plugin_filename, 'readonly': readonly,
-                    'info': info, 'components': []
-                }
-            plugins[dist.project_name]['components'].append({
-                'name': component.__name__, 'module': module.__name__,
-                'description': description,
-                'enabled': self.env.is_component_enabled(component),
-                'required': component.__name__ in self.required_components,
-            })
-
-        def component_order(a, b):
-            c = cmp(len(a['module'].split('.')), len(b['module'].split('.')))
-            if c == 0:
-                c = cmp(a['module'].lower(), b['module'].lower())
-                if c == 0:
-                    c = cmp(a['name'].lower(), b['name'].lower())
-            return c
-        for category in plugins:
-            plugins[category]['components'].sort(component_order)
-
-        plugin_list = [plugins['Trac']]
-        addons = [key for key in plugins.keys() if key != 'Trac']
-        addons.sort()
-        plugin_list += [plugins[category] for category in addons]
+        def safe_wiki_to_html(context, text):
+            try:
+                return format_to_html(self.env, context, text)
+            except Exception, e:
+                self.log.error('Unable to render component documentation: %s',
+                               exception_to_unicode(e, traceback=True))
+                return tag.pre(text)
 
         data = {
-            'plugins': plugin_list,
-            'readonly': not os.access(plugins_dir, os.F_OK + os.W_OK)
+            'plugins': plugins, 'show': req.args.get('show'),
+            'readonly': not os.access(get_plugins_dir(self.env),
+                                      os.F_OK + os.W_OK),
+            'safe_wiki_to_html': safe_wiki_to_html,
         }
         return 'admin_plugins.html', data
-
-    def _find_distribution(self, module):
-        path = get_module_path(module)
-        if path == self.trac_path:
-            return pkg_resources.Distribution(project_name='Trac',
-                                              version=TRAC_VERSION,
-                                              location=path)
-        for dist in pkg_resources.find_distributions(path, only=True):
-            return dist
-        else:
-            # This is a plain Python source file, not an egg
-            return pkg_resources.Distribution(project_name=module.__name__,
-                                              version='',
-                                              location=module.__file__)
