@@ -14,12 +14,18 @@
 # individuals. For the exact contribution history, see the revision
 # history and logs, available at http://trac.edgewall.org/log/.
 
-import re
+import os, re, types
+
+from genshi.core import Markup
 
 from trac.core import *
-from trac.db.api import IDatabaseConnector
-from trac.db.util import ConnectionWrapper
+from trac.config import Option
+from trac.db.api import IDatabaseConnector, _parse_db_str
+from trac.db.util import ConnectionWrapper, IterableCursor
 from trac.util import get_pkginfo
+from trac.util.compat import close_fds
+from trac.util.text import to_unicode
+from trac.util.translation import _
 
 _like_escape_re = re.compile(r'([/_%])')
 
@@ -46,28 +52,37 @@ try:
 except ImportError:
     has_mysqldb = False
 
-class MySQLConnector(Component):
-    """MySQL database support for version 4.1 and greater.
-    
-    Database urls should be of the form:
-        mysql://user[:password]@host[:port]/database
-    """
+# Mapping from "abstract" SQL types to DB-specific types
+_type_map = {
+    'int64': 'bigint',
+}
 
+
+class MySQLConnector(Component):
+    """Database connector for MySQL version 4.1 and greater.
+    
+    Database URLs should be of the form:
+    {{{
+    mysql://user[:password]@host[:port]/database
+    }}}
+    """
     implements(IDatabaseConnector)
 
+    mysqldump_path = Option('trac', 'mysqldump_path', 'mysqldump',
+        """Location of mysqldump for MySQL database backups""")
+
     def __init__(self):
         self._version = None
+        self.error = None
 
     def get_supported_schemes(self):
-        global has_mysqldb
-        if has_mysqldb:
-            return [('mysql', 1)]
-        else:
-            return []
+        if not has_mysqldb:
+            self.error = _("Cannot load Python bindings for MySQL")
+        yield ('mysql', self.error and -1 or 1)
 
-    def get_connection(self, path, user=None, password=None, host=None,
-                       port=None, params={}):
-        cnx = MySQLConnection(path, user, password, host, port, params)
+    def get_connection(self, path, log=None, user=None, password=None,
+                       host=None, port=None, params={}):
+        cnx = MySQLConnection(path, log, user, password, host, port, params)
         if not self._version:
             self._version = get_pkginfo(MySQLdb).get('version',
                                                      MySQLdb.__version__)
@@ -77,11 +92,13 @@ class MySQLConnector(Component):
                           MySQLdb.thread_safe())
             self.env.systeminfo.extend([('MySQL', mysql_info),
                                         ('MySQLdb', self._version)])
+            self.required = True
         return cnx
     
-    def init_db(self, path, user=None, password=None, host=None, port=None,
-                params={}):
-        cnx = self.get_connection(path, user, password, host, port, params)
+    def init_db(self, path, log=None, user=None, password=None, host=None,
+                port=None, params={}):
+        cnx = self.get_connection(path, log, user, password, host, port,
+                                  params)
         cursor = cnx.cursor()
         from trac.db_default import schema
         for table in schema:
@@ -105,7 +122,9 @@ class MySQLConnector(Component):
             name = '`%s`' % c
             table_col = filter((lambda x: x.name == c), table.columns)
             if len(table_col) == 1 and table_col[0].type.lower() == 'text':
-                if name == '`rev`':
+                if table_col[0].key_size is not None:
+                    name += '(%d)' % table_col[0].key_size
+                elif name == '`rev`':
                     name += '(20)'
                 elif name == '`path`':
                     name += '(255)'
@@ -123,6 +142,7 @@ class MySQLConnector(Component):
         coldefs = []
         for column in table.columns:
             ctype = column.type
+            ctype = _type_map.get(ctype, ctype)
             if column.auto_increment:
                 ctype = 'INT UNSIGNED NOT NULL AUTO_INCREMENT'
                 # Override the column type, as a text field cannot
@@ -141,13 +161,59 @@ class MySQLConnector(Component):
                   '_'.join(index.columns), table.name,
                   self._collist(table, index.columns))
 
+    def alter_column_types(self, table, columns):
+        """Yield SQL statements altering the type of one or more columns of
+        a table.
+        
+        Type changes are specified as a `columns` dict mapping column names
+        to `(from, to)` SQL type tuples.
+        """
+        alterations = []
+        for name, (from_, to) in sorted(columns.iteritems()):
+            to = _type_map.get(to, to)
+            if to != _type_map.get(from_, from_):
+                alterations.append((name, to))
+        if alterations:
+            yield "ALTER TABLE %s %s" % (table,
+                ', '.join("MODIFY %s %s" % each
+                          for each in alterations))
+
+    def backup(self, dest_file):
+        try:
+            from subprocess import Popen, PIPE
+        except ImportError:
+            raise TracError('Python >= 2.4 or the subprocess module '
+                            'is required for pre-upgrade backup support')
+        db_url = self.env.config.get('trac', 'database')
+        scheme, db_prop = _parse_db_str(db_url)
+        db_name = os.path.basename(db_prop['path'])
+
+        args = [self.mysqldump_path]
+        if 'host' in db_prop:
+            args.extend(['-h', db_prop['host']])
+        if 'port' in db_prop:
+            args.extend(['-P', str(db_prop['port'])])
+        if 'user' in db_prop:
+            args.extend(['-u', db_prop['user']])
+        args.extend(['-r', dest_file, db_name])
+        
+        environ = os.environ.copy()
+        environ['MYSQL_PWD'] = str(db_prop['password'])
+        p = Popen(args, env=environ, stderr=PIPE, close_fds=close_fds)
+        errmsg = p.communicate()[1]
+        if p.returncode != 0:
+            raise TracError("Backup attempt failed (%s)" % to_unicode(errmsg))
+        if not os.path.exists(dest_file):
+            raise TracError("Backup attempt failed")
+        return dest_file
+
 
 class MySQLConnection(ConnectionWrapper):
     """Connection wrapper for MySQL."""
 
     poolable = True
 
-    def __init__(self, path, user=None, password=None, host=None,
+    def __init__(self, path, log, user=None, password=None, host=None,
                  port=None, params={}):
         if path.startswith('/'):
             path = path[1:]
@@ -157,11 +223,14 @@ class MySQLConnection(ConnectionWrapper):
             port = 3306
         cnx = MySQLdb.connect(db=path, user=user, passwd=password,
                               host=host, port=port, charset='utf8')
-        ConnectionWrapper.__init__(self, cnx)
+        if hasattr(cnx, 'encoders'):
+            # 'encoders' undocumented but present since 1.2.1 (r422)
+            cnx.encoders[Markup] = cnx.encoders[types.UnicodeType]
+        ConnectionWrapper.__init__(self, cnx, log)
         self._is_closed = False
 
     def cast(self, column, type):
-        if type == 'int':
+        if type == 'int'or type == 'int64':
             type = 'signed'
         elif type == 'text':
             type = 'char'
@@ -171,13 +240,18 @@ class MySQLConnection(ConnectionWrapper):
         return 'concat(%s)' % ', '.join(args)
 
     def like(self):
-        return "LIKE %s ESCAPE '/'"
+        """Return a case-insensitive LIKE clause."""
+        return "LIKE %s COLLATE utf8_general_ci ESCAPE '/'"
 
     def like_escape(self, text):
         return _like_escape_re.sub(r'/\1', text)
 
+    def quote(self, identifier):
+        """Return the quoted identifier."""
+        return "`%s`" % identifier
+
     def get_last_id(self, cursor, table, column='id'):
-        return self.cnx.insert_id()
+        return cursor.lastrowid
 
     def rollback(self):
         self.cnx.ping()
@@ -195,4 +269,4 @@ class MySQLConnection(ConnectionWrapper):
             self._is_closed = True
 
     def cursor(self):
-        return MySQLUnicodeCursor(self.cnx)
+        return IterableCursor(MySQLUnicodeCursor(self.cnx), self.log)
