@@ -16,13 +16,13 @@
 # Author: Jonas Borgström <jonas@edgewall.com>
 #         Christopher Lenz <cmlenz@gmx.de>
 
-"""Management of permissions."""
-
 from time import time
+
+from trac.admin import AdminCommandError, IAdminCommandProvider
 from trac.config import ExtensionOption, OrderedExtensionsOption
 from trac.core import *
 from trac.resource import Resource, get_resource_name
-from trac.util.compat import set
+from trac.util.text import print_table, printout, wrap
 from trac.util.translation import _
 
 __all__ = ['IPermissionRequestor', 'IPermissionStore',
@@ -150,9 +150,9 @@ class IPermissionPolicy(Interface):
 
 
 class DefaultPermissionStore(Component):
-    """Default implementation of permission storage and simple group management.
+    """Default implementation of permission storage and group management.
     
-    This component uses the `PERMISSION` table in the database to store both
+    This component uses the `permission` table in the database to store both
     permissions and groups.
     """
     implements(IPermissionStore)
@@ -170,7 +170,7 @@ class DefaultPermissionStore(Component):
         """
         subjects = set([username])
         for provider in self.group_providers:
-            subjects.update(provider.get_permission_groups(username))
+            subjects.update(provider.get_permission_groups(username) or [])
 
         actions = set([])
         db = self.env.get_db_cnx()
@@ -200,8 +200,6 @@ class DefaultPermissionStore(Component):
         # get_user_permissions() takes care of the magic 'authenticated' group.
         # The optimized loop we had before didn't.  This is very inefficient,
         # but it works.
-        db = self.env.get_db_cnx()
-        cursor = db.cursor()
         result = set()
         users = set([u[0] for u in self.env.get_known_users()])
         for user in users:
@@ -223,26 +221,29 @@ class DefaultPermissionStore(Component):
 
     def grant_permission(self, username, action):
         """Grants a user the permission to perform the specified action."""
-        db = self.env.get_db_cnx()
-        cursor = db.cursor()
-        cursor.execute("INSERT INTO permission VALUES (%s, %s)",
-                       (username, action))
+        @self.env.with_transaction()
+        def do_grant(db):
+            cursor = db.cursor()
+            cursor.execute("INSERT INTO permission VALUES (%s, %s)",
+                           (username, action))
         self.log.info('Granted permission for %s to %s' % (action, username))
-        db.commit()
 
     def revoke_permission(self, username, action):
         """Revokes a users' permission to perform the specified action."""
-        db = self.env.get_db_cnx()
-        cursor = db.cursor()
-        cursor.execute("DELETE FROM permission WHERE username=%s AND action=%s",
-                       (username, action))
+        @self.env.with_transaction()
+        def do_revoke(db):
+            cursor = db.cursor()
+            cursor.execute("DELETE FROM permission WHERE username=%s "
+                           "AND action=%s",
+                           (username, action))
         self.log.info('Revoked permission for %s to %s' % (action, username))
-        db.commit()
 
 
 class DefaultPermissionGroupProvider(Component):
-    """Provides the basic builtin permission groups 'anonymous' and
-    'authenticated'."""
+    """Permission group provider providing the basic builtin permission groups
+    'anonymous' and 'authenticated'."""
+
+    required = True
 
     implements(IPermissionGroupProvider)
 
@@ -290,7 +291,9 @@ class DefaultPermissionPolicy(Component):
 
 
 class PermissionSystem(Component):
-    """Sub-system that manages user permissions."""
+    """Permission management sub-system."""
+
+    required = True
 
     implements(IPermissionRequestor)
 
@@ -312,6 +315,15 @@ class PermissionSystem(Component):
         LegacyAttachmentPolicy (map ATTACHMENT_* permissions to realm specific
         ones)""")
 
+    # Number of seconds a cached user permission set is valid for.
+    CACHE_EXPIRY = 5
+    # How frequently to clear the entire permission cache
+    CACHE_REAP_TIME = 60
+
+    def __init__(self):
+        self.permission_cache = {}
+        self.last_reap = time()
+
     # Public API
 
     def grant_permission(self, username, action):
@@ -329,7 +341,7 @@ class PermissionSystem(Component):
     def get_actions(self):
         actions = []
         for requestor in self.requestors:
-            for action in requestor.get_permission_actions():
+            for action in requestor.get_permission_actions() or []:
                 if isinstance(action, tuple):
                     actions.append(action[0])
                 else:
@@ -344,7 +356,7 @@ class PermissionSystem(Component):
         means the permission is denied."""
         actions = []
         for requestor in self.requestors:
-            actions += list(requestor.get_permission_actions())
+            actions += list(requestor.get_permission_actions() or [])
         permissions = {}
         if username:
             # Return all permissions that the given user has
@@ -357,7 +369,7 @@ class PermissionSystem(Component):
                 permissions[action] = True
                 if meta.has_key(action):
                     [_expand_meta(perm) for perm in meta[action]]
-            for perm in self.store.get_user_permissions(username):
+            for perm in self.store.get_user_permissions(username) or []:
                 _expand_meta(perm)
         else:
             # Return all permissions available in the system
@@ -373,17 +385,27 @@ class PermissionSystem(Component):
 
         The permissions are returned as a list of (subject, action)
         formatted tuples."""
-        return self.store.get_all_permissions()
+        return self.store.get_all_permissions() or []
 
     def get_users_with_permission(self, permission):
         """Return all users that have the specified permission.
         
         Users are returned as a list of user names.
         """
-        # this should probably be cached
+        now = time()
+
+        if now - self.last_reap > self.CACHE_REAP_TIME:
+            self.permission_cache = {}
+            self.last_reap = time()
+
+        timestamp, permissions = self.permission_cache.get(permission, 
+                                                           (0, None))
+        if now - timestamp <= self.CACHE_EXPIRY:
+            return permissions
+
         parent_map = {}
         for requestor in self.requestors:
-            for action in requestor.get_permission_actions():
+            for action in requestor.get_permission_actions() or []:
                 for child in action[1]:
                     parent_map.setdefault(child, []).append(action[0])
 
@@ -396,14 +418,18 @@ class PermissionSystem(Component):
                 map(_append_with_parents, parent_map[action])
         _append_with_parents(permission)
 
-        return self.store.get_users_with_permissions(satisfying_perms.keys())
+        perms = self.store.get_users_with_permissions(satisfying_perms.keys())
+        perms = perms or []
+        self.permission_cache[permission] = (now, perms)
+
+        return perms
 
     def expand_actions(self, actions):
         """Helper method for expanding all meta actions."""
         actions = list(actions)     # Consume actions if it is an iterator
         meta = {}
         for requestor in self.requestors:
-            for m in requestor.get_permission_actions():
+            for m in requestor.get_permission_actions() or []:
                 if isinstance(m, tuple):
                     meta[m[0]] = m[1]
         expanded_actions = set(actions)
@@ -447,7 +473,7 @@ class PermissionSystem(Component):
         """
         actions = ['EMAIL_VIEW']
         for requestor in [r for r in self.requestors if r is not self]:
-            for action in requestor.get_permission_actions():
+            for action in requestor.get_permission_actions() or []:
                 if isinstance(action, tuple):
                     actions.append(action[0])
                 else:
@@ -557,3 +583,94 @@ class PermissionCache(object):
         perm = PermissionSystem(self.env)
         actions = perm.get_user_permissions(self.username)
         return [action for action in actions if action in self]
+
+
+class PermissionAdmin(Component):
+    """trac-admin command provider for permission system administration."""
+    
+    implements(IAdminCommandProvider)
+    
+    # IAdminCommandProvider methods
+    
+    def get_admin_commands(self):
+        yield ('permission list', '[user]',
+               'List permission rules',
+               self._complete_list, self._do_list)
+        yield ('permission add', '<user> <action> [action] [...]',
+               'Add a new permission rule',
+               self._complete_add, self._do_add)
+        yield ('permission remove', '<user> <action> [action] [...]',
+               'Remove a permission rule',
+               self._complete_remove, self._do_remove)
+    
+    def get_user_list(self):
+        return set(user for (user, action) in 
+                   PermissionSystem(self.env).get_all_permissions())
+    
+    def get_user_perms(self, user):
+        return [action for (subject, action) in
+                PermissionSystem(self.env).get_all_permissions()
+                if subject == user]
+    
+    def _complete_list(self, args):
+        if len(args) == 1:
+            return self.get_user_list()
+    
+    def _complete_add(self, args):
+        if len(args) == 1:
+            return self.get_user_list()
+        elif len(args) >= 2:
+            return (set(PermissionSystem(self.env).get_actions())
+                    - set(self.get_user_perms(args[0])) - set(args[1:-1]))
+    
+    def _complete_remove(self, args):
+        if len(args) == 1:
+            return self.get_user_list()
+        elif len(args) >= 2:
+            return set(self.get_user_perms(args[0])) - set(args[1:-1])
+    
+    def _do_list(self, user=None):
+        permsys = PermissionSystem(self.env)
+        if user:
+            rows = []
+            perms = permsys.get_user_permissions(user)
+            for action in perms:
+                if perms[action]:
+                    rows.append((user, action))
+        else:
+            rows = permsys.get_all_permissions()
+        rows.sort()
+        print_table(rows, [_('User'), _('Action')])
+        print
+        printout(_("Available actions:"))
+        actions = permsys.get_actions()
+        actions.sort()
+        text = ', '.join(actions)
+        printout(wrap(text, initial_indent=' ', subsequent_indent=' ', 
+                      linesep='\n'))
+        print
+    
+    def _do_add(self, user, *actions):
+        permsys = PermissionSystem(self.env)
+        if user.isupper():
+            raise AdminCommandError(_('All upper-cased tokens are reserved '
+                                      'for permission names'))
+        for action in actions:
+            permsys.grant_permission(user, action)
+    
+    def _do_remove(self, user, *actions):
+        permsys = PermissionSystem(self.env)
+        rows = permsys.get_all_permissions()
+        for action in actions:
+            if action == '*':
+                for row in rows:
+                    if user != '*' and user != row[0]:
+                        continue
+                    permsys.revoke_permission(row[0], row[1])
+            else:
+                for row in rows:
+                    if action != row[1]:
+                        continue
+                    if user != '*' and user != row[0]:
+                        continue
+                    permsys.revoke_permission(row[0], row[1])
