@@ -17,20 +17,25 @@
 # Author: Jonas Borgström <jonas@edgewall.com>
 #         Christopher Lenz <cmlenz@gmx.de>
 
+import doctest
 import os
 import unittest
 import sys
-import pkg_resources
-from fnmatch import fnmatch
+
+try:
+    from babel import Locale
+except ImportError:
+    Locale = None
 
 from trac.config import Configuration
-from trac.core import Component, ComponentManager, ExtensionPoint
+from trac.core import Component, ComponentManager
 from trac.env import Environment
 from trac.db.api import _parse_db_str, DatabaseManager
 from trac.db.sqlite_backend import SQLiteConnection
 import trac.db.postgres_backend
 import trac.db.mysql_backend
 from trac.ticket.default_workflow import load_workflow_config_snippet
+from trac.util import translation
 
 
 def Mock(bases=(), *initargs, **kw):
@@ -85,7 +90,7 @@ def Mock(bases=(), *initargs, **kw):
         bases = (bases,)
     cls = type('Mock', bases, {})
     mock = cls(*initargs)
-    for k,v in kw.items():
+    for k, v in kw.items():
         setattr(mock, k, v)
     return mock
 
@@ -93,6 +98,9 @@ def Mock(bases=(), *initargs, **kw):
 class MockPerm(object):
     """Fake permission class. Necessary as Mock can not be used with operator
     overloading."""
+
+    username = ''
+    
     def has_permission(self, action, realm_or_resource=None, id=False,
                        version=False):
         return True
@@ -127,16 +135,20 @@ class TestSetup(unittest.TestSuite):
             for test in self._tests:
                 if hasattr(test, 'setFixture'):
                     test.setFixture(self.fixture)
-        for test in self._tests:    # Content of unittest.TestSuite.run()
-            if result.shouldStop:   # copied here for Python 2.3 compatibility
-                break
-            test(result)
+        unittest.TestSuite.run(self, result)
         self.tearDown()
         return result
 
-    def __call__(self, *args, **kwds):      # Python 2.3 compatibility
-        return self.run(*args, **kwds)
-
+    def _wrapped_run(self, result):
+        "Python 2.7 / unittest2 compatibility - there must be a better way..."
+        self.setUp()
+        if hasattr(self, 'fixture'):
+            for test in self._tests:
+                if hasattr(test, 'setFixture'):
+                    test.setFixture(self.fixture)
+        unittest.TestSuite._wrapped_run(self, result)
+        self.tearDown()
+        return result
 
 class TestCaseSetup(unittest.TestCase):
     def setFixture(self, fixture):
@@ -233,8 +245,7 @@ class EnvironmentStub(Environment):
         """
         ComponentManager.__init__(self)
         Component.__init__(self)
-        self.enabled_components = enable or ['trac.*']
-        self.systeminfo = [('Python', sys.version)]
+        self.systeminfo = []
 
         import trac
         self.path = os.path.dirname(trac.__file__)
@@ -249,10 +260,15 @@ class EnvironmentStub(Environment):
         load_workflow_config_snippet(self.config, 'basic-workflow.ini')
         self.config.set('logging', 'log_level', 'DEBUG')
         self.config.set('logging', 'log_type', 'stderr')
+        if enable is not None:
+            self.config.set('components', 'trac.*', 'disabled')
+        for name_or_class in enable or ():
+            config_key = self._component_name(name_or_class)
+            self.config.set('components', config_key, 'enabled')
 
         # -- logging
-        from trac.log import logger_factory
-        self.log = logger_factory('test')
+        from trac.log import logger_handler_factory
+        self.log, self._log_handler = logger_handler_factory('test')
 
         # -- database
         self.dburi = get_dburi()
@@ -267,16 +283,11 @@ class EnvironmentStub(Environment):
         self.abs_href = Href('http://example.org/trac.cgi')
 
         self.known_users = []
+        translation.activate(Locale and Locale('en', 'US'))
 
-    def is_component_enabled(self, cls):
-        for component in self.enabled_components:
-            if component is cls:
-                return True
-            if isinstance(component, basestring) and \
-                fnmatch(cls.__module__ + '.' + cls.__name__, component):
-                return True
-        return False
-
+    def get_read_db(self):
+        return self.get_db_cnx()
+    
     def get_db_cnx(self, destroying=False):
         if self.db:
             return self.db # in-memory SQLite
@@ -368,7 +379,7 @@ class EnvironmentStub(Environment):
                 for t in tables:
                     cursor.execute('DROP TABLE IF EXISTS `%s`' % t)
             db.commit()
-        except Exception, e:
+        except Exception:
             db.rollback()
 
     def get_known_users(self, cnx=None):
@@ -380,7 +391,6 @@ def locate(fn):
 
     Returns the fully-qualified path, or None.
     """
-    import os
     exec_suffix = os.name == 'nt' and '.exe' or ''
     
     for p in ["."] + os.environ['PATH'].split(os.pathsep):
@@ -403,6 +413,7 @@ def suite():
     import trac.versioncontrol.web_ui.tests
     import trac.web.tests
     import trac.wiki.tests
+    import tracopt.mimeview.tests
 
     suite = unittest.TestSuite()
     suite.addTest(trac.tests.basicSuite())
@@ -417,19 +428,12 @@ def suite():
     suite.addTest(trac.versioncontrol.web_ui.tests.suite())
     suite.addTest(trac.web.tests.suite())
     suite.addTest(trac.wiki.tests.suite())
+    suite.addTest(tracopt.mimeview.tests.suite())
+    suite.addTest(doctest.DocTestSuite(sys.modules[__name__]))
 
     return suite
 
 if __name__ == '__main__':
-    import doctest, sys
-    doctest.testmod(sys.modules[__name__])
-
-    # Clean up after doctest or spambayes gets unhappy
-    try:
-        del __builtins__._
-    except NameError:
-        pass
-
     #FIXME: this is a bit inelegant
     if '--skip-functional-tests' in sys.argv:
         sys.argv.remove('--skip-functional-tests')
