@@ -14,24 +14,19 @@
 #
 # Author: Jonas Borgström <jonas@edgewall.com>
 
+import copy
 import re
-from datetime import datetime
-try:
-    import threading
-except ImportError:
-    import dummy_threading as threading
 
 from genshi.builder import tag
 
+from trac.cache import cached
 from trac.config import *
 from trac.core import *
 from trac.perm import IPermissionRequestor, PermissionCache, PermissionSystem
 from trac.resource import IResourceManager
 from trac.util import Ranges
-from trac.util.compat import set, sorted
-from trac.util.datefmt import utc
-from trac.util.text import shorten_line, obfuscate_email_address
-from trac.util.translation import _
+from trac.util.text import shorten_line
+from trac.util.translation import _, N_, gettext
 from trac.wiki import IWikiSyntaxProvider, WikiParser
 
 
@@ -143,10 +138,31 @@ class ITicketManipulator(Interface):
         ticket. Therefore, a return value of `[]` means everything is OK."""
 
 
+class IMilestoneChangeListener(Interface):
+    """Extension point interface for components that require notification
+    when milestones are created, modified, or deleted."""
+
+    def milestone_created(milestone):
+        """Called when a milestone is created."""
+
+    def milestone_changed(milestone, old_values):
+        """Called when a milestone is modified.
+
+        `old_values` is a dictionary containing the previous values of the
+        milestone properties that changed. Currently those properties can be
+        'name', 'due', 'completed', or 'description'.
+        """
+
+    def milestone_deleted(milestone):
+        """Called when a milestone is deleted."""
+
+
 class TicketSystem(Component):
     implements(IPermissionRequestor, IWikiSyntaxProvider, IResourceManager)
 
     change_listeners = ExtensionPoint(ITicketChangeListener)
+    milestone_change_listeners = ExtensionPoint(IMilestoneChangeListener)
+    
     action_controllers = OrderedExtensionsOption('ticket', 'workflow',
         ITicketActionController, default='ConfigurableTicketWorkflow',
         include_missing=False,
@@ -154,17 +170,52 @@ class TicketSystem(Component):
             (''since 0.11'').""")
 
     restrict_owner = BoolOption('ticket', 'restrict_owner', 'false',
-        """Make the owner field of tickets use a drop-down menu. See
+        """Make the owner field of tickets use a drop-down menu.
+        Be sure to understand the performance implications before activating
+        this option. See
         [TracTickets#Assign-toasDrop-DownList Assign-to as Drop-Down List]
         (''since 0.9'').""")
 
-    _fields = None
-    _custom_fields = None
+    default_version = Option('ticket', 'default_version', '',
+        """Default version for newly created tickets.""")
+
+    default_type = Option('ticket', 'default_type', 'defect',
+        """Default type for newly created tickets (''since 0.9'').""")
+
+    default_priority = Option('ticket', 'default_priority', 'major',
+        """Default priority for newly created tickets.""")
+
+    default_milestone = Option('ticket', 'default_milestone', '',
+        """Default milestone for newly created tickets.""")
+
+    default_component = Option('ticket', 'default_component', '',
+        """Default component for newly created tickets.""")
+
+    default_severity = Option('ticket', 'default_severity', '',
+        """Default severity for newly created tickets.""")
+
+    default_summary = Option('ticket', 'default_summary', '',
+        """Default summary (title) for newly created tickets.""")
+
+    default_description = Option('ticket', 'default_description', '',
+        """Default description for newly created tickets.""")
+
+    default_keywords = Option('ticket', 'default_keywords', '',
+        """Default keywords for newly created tickets.""")
+
+    default_owner = Option('ticket', 'default_owner', '',
+        """Default owner for newly created tickets.""")
+
+    default_cc = Option('ticket', 'default_cc', '',
+        """Default cc: list for newly created tickets.""")
+
+    default_resolution = Option('ticket', 'default_resolution', 'fixed',
+        """Default resolution for resolving (closing) tickets
+        (''since 0.11'').""")
 
     def __init__(self):
         self.log.debug('action controllers for ticket workflow: %r' % 
                 [c.__class__.__name__ for c in self.action_controllers])
-        self._fields_lock = threading.RLock()
 
     # Public API
 
@@ -173,7 +224,7 @@ class TicketSystem(Component):
         # The list should not have duplicates.
         actions = {}
         for controller in self.action_controllers:
-            weighted_actions = controller.get_ticket_actions(req, ticket)
+            weighted_actions = controller.get_ticket_actions(req, ticket) or []
             for weight, action in weighted_actions:
                 if action in actions:
                     actions[action] = max(actions[action], weight)
@@ -188,68 +239,72 @@ class TicketSystem(Component):
         controllers know about."""
         valid_states = set()
         for controller in self.action_controllers:
-            valid_states.update(controller.get_all_status())
+            valid_states.update(controller.get_all_status() or [])
         return sorted(valid_states)
 
+    def get_ticket_field_labels(self):
+        """Produce a (name,label) mapping from `get_ticket_fields`."""
+        return dict((f['name'], f['label']) for f in
+                    TicketSystem(self.env).get_ticket_fields())
+
     def get_ticket_fields(self):
-        """Returns the list of fields available for tickets."""
-        # This is now cached - as it makes quite a number of things faster,
-        # e.g. #6436
-        if self._fields is None:
-            self._fields_lock.acquire()
-            try:
-                if self._fields is None: # double-check (race after 1st check)
-                    self._fields = self._get_ticket_fields()
-            finally:
-                self._fields_lock.release()
-        return [f.copy() for f in self._fields]
+        """Returns list of fields available for tickets.
+
+        Each field is a dict with at least the 'name', 'label' (localized)
+        and 'type' keys.
+        It may in addition contain the 'custom' key, the 'optional' and the
+        'options' keys. When present 'custom' and 'optional' are always `True`.
+        """
+        fields = copy.deepcopy(self.fields)
+        label = 'label' # workaround gettext extraction bug
+        for f in fields:
+            f[label] = gettext(f[label])
+        return fields
 
     def reset_ticket_fields(self):
-        self._fields_lock.acquire()
-        try:
-            self._fields = None
-            self.config.touch() # brute force approach for now
-        finally:
-            self._fields_lock.release()
+        """Invalidate ticket field cache."""
+        del self.fields
 
-    def _get_ticket_fields(self):
+    @cached
+    def fields(self, db):
+        """Return the list of fields available for tickets."""
         from trac.ticket import model
 
-        db = self.env.get_db_cnx()
         fields = []
 
         # Basic text fields
-        for name in ('summary', 'reporter'):
-            field = {'name': name, 'type': 'text', 'label': name.title()}
-            fields.append(field)
+        fields.append({'name': 'summary', 'type': 'text',
+                       'label': N_('Summary')})
+        fields.append({'name': 'reporter', 'type': 'text',
+                       'label': N_('Reporter')})
 
         # Owner field, by default text but can be changed dynamically 
         # into a drop-down depending on configuration (restrict_owner=true)
-        field = {'name': 'owner', 'label': 'Owner'}
+        field = {'name': 'owner', 'label': N_('Owner')}
         field['type'] = 'text'
         fields.append(field)
 
         # Description
         fields.append({'name': 'description', 'type': 'textarea',
-                       'label': 'Description'})
+                       'label': N_('Description')})
 
         # Default select and radio fields
-        selects = [('type', model.Type),
-                   ('status', model.Status),
-                   ('priority', model.Priority),
-                   ('milestone', model.Milestone),
-                   ('component', model.Component),
-                   ('version', model.Version),
-                   ('severity', model.Severity),
-                   ('resolution', model.Resolution)]
-        for name, cls in selects:
+        selects = [('type', N_('Type'), model.Type),
+                   ('status', N_('Status'), model.Status),
+                   ('priority', N_('Priority'), model.Priority),
+                   ('milestone', N_('Milestone'), model.Milestone),
+                   ('component', N_('Component'), model.Component),
+                   ('version', N_('Version'), model.Version),
+                   ('severity', N_('Severity'), model.Severity),
+                   ('resolution', N_('Resolution'), model.Resolution)]
+        for name, label, cls in selects:
             options = [val.name for val in cls.select(self.env, db=db)]
             if not options:
                 # Fields without possible values are treated as if they didn't
                 # exist
                 continue
-            field = {'name': name, 'type': 'select', 'label': name.title(),
-                     'value': self.config.get('ticket', 'default_' + name),
+            field = {'name': name, 'type': 'select', 'label': label,
+                     'value': getattr(self, 'default_' + name, ''),
                      'options': options}
             if name in ('status', 'resolution'):
                 field['type'] = 'radio'
@@ -259,9 +314,15 @@ class TicketSystem(Component):
             fields.append(field)
 
         # Advanced text fields
-        for name in ('keywords', 'cc', ):
-            field = {'name': name, 'type': 'text', 'label': name.title()}
-            fields.append(field)
+        fields.append({'name': 'keywords', 'type': 'text',
+                       'label': N_('Keywords')})
+        fields.append({'name': 'cc', 'type': 'text', 'label': N_('Cc')})
+
+        # Date/time fields
+        fields.append({'name': 'time', 'type': 'time',
+                       'label': N_('Created')})
+        fields.append({'name': 'changetime', 'type': 'time',
+                       'label': N_('Modified')})
 
         for field in self.get_custom_fields():
             if field['name'] in [f['name'] for f in fields]:
@@ -282,19 +343,15 @@ class TicketSystem(Component):
         return fields
 
     reserved_field_names = ['report', 'order', 'desc', 'group', 'groupdesc',
-                            'col', 'row', 'format', 'max', 'page', 'verbose']
+                            'col', 'row', 'format', 'max', 'page', 'verbose',
+                            'comment', 'or']
 
     def get_custom_fields(self):
-        if self._custom_fields is None:
-            self._fields_lock.acquire()
-            try:
-                if self._custom_fields is None: # double-check
-                    self._custom_fields = self._get_custom_fields()
-            finally:
-                self._fields_lock.release()
-        return [f.copy() for f in self._custom_fields]
+        return copy.deepcopy(self.custom_fields)
 
-    def _get_custom_fields(self):
+    @cached
+    def custom_fields(self, db):
+        """Return the list of custom ticket fields available for tickets."""
         fields = []
         config = self.config['ticket-custom']
         for name in [option for option, value in config.options()
@@ -322,6 +379,12 @@ class TicketSystem(Component):
         fields.sort(lambda x, y: cmp(x['order'], y['order']))
         return fields
 
+    def get_field_synonyms(self):
+        """Return a mapping from field name synonyms to field names.
+        The synonyms are supposed to be more intuitive for custom queries."""
+        # i18n TODO - translated keys
+        return {'created': 'time', 'modified': 'changetime'}
+
     def eventually_restrict_owner(self, field, ticket=None):
         """Restrict given owner field to be a list of users having
         the TICKET_MODIFY permission (for the given ticket)
@@ -344,10 +407,12 @@ class TicketSystem(Component):
     def get_permission_actions(self):
         return ['TICKET_APPEND', 'TICKET_CREATE', 'TICKET_CHGPROP',
                 'TICKET_VIEW', 'TICKET_EDIT_CC', 'TICKET_EDIT_DESCRIPTION',
+                'TICKET_EDIT_COMMENT',
                 ('TICKET_MODIFY', ['TICKET_APPEND', 'TICKET_CHGPROP']),
                 ('TICKET_ADMIN', ['TICKET_CREATE', 'TICKET_MODIFY',
                                   'TICKET_VIEW', 'TICKET_EDIT_CC',
-                                  'TICKET_EDIT_DESCRIPTION'])]
+                                  'TICKET_EDIT_DESCRIPTION',
+                                  'TICKET_EDIT_COMMENT'])]
 
     # IWikiSyntaxProvider methods
 
@@ -448,3 +513,33 @@ class TicketSystem(Component):
             return "%s (%s)" % (summary, status)
         else:
             return summary
+
+    def resource_exists(self, resource):
+        """
+        >>> from trac.test import EnvironmentStub
+        >>> from trac.resource import Resource, resource_exists
+        >>> env = EnvironmentStub()
+
+        >>> resource_exists(env, Resource('ticket', 123456))
+        False
+
+        >>> from trac.ticket.model import Ticket
+        >>> t = Ticket(env)
+        >>> int(t.insert())
+        1
+        >>> resource_exists(env, t.resource)
+        True
+        """
+        db = self.env.get_read_db()
+        cursor = db.cursor()
+        cursor.execute("SELECT id FROM ticket WHERE id=%s", (resource.id,))
+        latest_exists = bool(cursor.fetchall())
+        if latest_exists:
+            if resource.version is None:
+                return True
+            cursor.execute("""
+                SELECT count(distinct time) FROM ticket_change WHERE ticket=%s
+                """, (resource.id,))
+            return cursor.fetchone()[0] >= resource.version
+        else:
+            return False
