@@ -16,15 +16,16 @@
 # Author: Daniel Lundin <daniel@edgewall.com>
 #
 
-from trac import __version__
 from trac.core import *
 from trac.config import *
 from trac.notification import NotifyEmail
+from trac.ticket.api import TicketSystem
 from trac.util import md5
-from trac.util.datefmt import to_timestamp
-from trac.util.text import CRLF, wrap, to_unicode, obfuscate_email_address
+from trac.util.datefmt import to_utimestamp
+from trac.util.text import CRLF, wrap, obfuscate_email_address
+from trac.util.translation import deactivate, reactivate
 
-from genshi.template.text import TextTemplate
+from genshi.template.text import NewTextTemplate
 
 class TicketNotificationSystem(Component):
 
@@ -66,6 +67,17 @@ class TicketNotifyEmail(NotifyEmail):
         self.prev_cc = []
 
     def notify(self, ticket, newticket=True, modtime=None):
+        """Send ticket change notification e-mail (untranslated)"""
+        t = deactivate()
+        translated_fields = ticket.fields
+        try:
+            ticket.fields = TicketSystem(self.env).get_ticket_fields()
+            self._notify(ticket, newticket, modtime)
+        finally:
+            ticket.fields = translated_fields
+            reactivate(t)
+
+    def _notify(self, ticket, newticket=True, modtime=None):
         self.ticket = ticket
         self.modtime = modtime
         self.newticket = newticket
@@ -132,12 +144,14 @@ class TicketNotifyEmail(NotifyEmail):
                         changes_body += '  * %s:  %s%s' % (field, chg, CRLF)
                     if newv:
                         change_data[field] = {'oldvalue': old, 'newvalue': new}
-            
-        self.ticket['description'] = wrap(
-            self.ticket.values.get('description', ''), self.COLS,
+        
+        ticket_values = ticket.values.copy()
+        ticket_values['id'] = ticket.id
+        ticket_values['description'] = wrap(
+            ticket_values.get('description', ''), self.COLS,
             initial_indent=' ', subsequent_indent=' ', linesep=CRLF)
-        self.ticket['new'] = self.newticket
-        self.ticket['link'] = link
+        ticket_values['new'] = self.newticket
+        ticket_values['link'] = link
         
         subject = self.format_subj(summary)
         if not self.newticket:
@@ -146,7 +160,7 @@ class TicketNotifyEmail(NotifyEmail):
             'ticket_props': self.format_props(),
             'ticket_body_hdr': self.format_hdr(),
             'subject': subject,
-            'ticket': ticket.values,
+            'ticket': ticket_values,
             'changes_body': changes_body,
             'changes_descr': changes_descr,
             'change': change_data
@@ -155,23 +169,27 @@ class TicketNotifyEmail(NotifyEmail):
 
     def format_props(self):
         tkt = self.ticket
-        fields = [f for f in tkt.fields if f['name'] not in ('summary', 'cc')]
+        fields = [f for f in tkt.fields 
+                  if f['name'] not in ('summary', 'cc', 'time', 'changetime')]
         width = [0, 0, 0, 0]
         i = 0
-        for f in [f['name'] for f in fields if f['type'] != 'textarea']:
-            if not tkt.values.has_key(f):
+        for f in fields:
+            if f['type'] == 'textarea':
                 continue
-            fval = tkt[f] or ''
+            fname = f['name']
+            if not fname in tkt.values:
+                continue
+            fval = tkt[fname] or ''
             if fval.find('\n') != -1:
                 continue
             idx = 2 * (i % 2)
-            if len(f) > width[idx]:
-                width[idx] = len(f)
+            if len(f['label']) > width[idx]:
+                width[idx] = len(f['label'])
             if len(fval) > width[idx + 1]:
                 width[idx + 1] = len(fval)
             i += 1
-        format = ('%%%is:  %%-%is  |  ' % (width[0], width[1]),
-                  ' %%%is:  %%-%is%s' % (width[2], width[3], CRLF))
+        format = (u'%%%is:  %%-%is  |  ' % (width[0], width[1]),
+                  u' %%%is:  %%-%is%s' % (width[2], width[3], CRLF))
         l = (width[0] + width[1] + 5)
         sep = l * '-' + '+' + (self.COLS - l) * '-'
         txt = sep + CRLF
@@ -185,9 +203,11 @@ class TicketNotifyEmail(NotifyEmail):
             if fname in ['owner', 'reporter']:
                 fval = obfuscate_email_address(fval)
             if f['type'] == 'textarea' or '\n' in unicode(fval):
-                big.append((fname.capitalize(), CRLF.join(fval.splitlines())))
+                big.append((f['label'], CRLF.join(fval.splitlines())))
             else:
-                txt += format[i % 2] % (fname.capitalize(), fval)
+                # Note: f['label'] is a Babel's LazyObject, make sure its
+                # __str__ method won't be called.
+                txt += format[i % 2] % (f['label'], unicode(fval))
                 i += 1
         if i % 2:
             txt += CRLF
@@ -216,11 +236,11 @@ class TicketNotifyEmail(NotifyEmail):
 
     def format_subj(self, summary):
         template = self.config.get('notification','ticket_subject_template')
-        template = TextTemplate(template.encode('utf8'))
+        template = NewTextTemplate(template.encode('utf8'))
                                                 
         prefix = self.config.get('notification', 'smtp_subject_prefix')
         if prefix == '__default__': 
-            prefix = '[%s]' % self.config.get('project', 'name')
+            prefix = '[%s]' % self.env.project_name
         
         data = {
             'prefix': prefix,
@@ -260,7 +280,7 @@ class TicketNotifyEmail(NotifyEmail):
         if notify_updater:
             cursor.execute("SELECT DISTINCT author,ticket FROM ticket_change "
                            "WHERE ticket=%s", (tktid,))
-            for author,ticket in cursor:
+            for author, ticket in cursor:
                 torecipients.append(author)
 
         # Suppress the updater from the recipients
@@ -290,8 +310,8 @@ class TicketNotifyEmail(NotifyEmail):
 
     def get_message_id(self, rcpt, modtime=None):
         """Generate a predictable, but sufficiently unique message ID."""
-        s = '%s.%08d.%d.%s' % (self.config.get('project', 'url'),
-                               int(self.ticket.id), to_timestamp(modtime),
+        s = '%s.%08d.%d.%s' % (self.env.project_url.encode('utf-8'),
+                               int(self.ticket.id), to_utimestamp(modtime),
                                rcpt.encode('ascii', 'ignore'))
         dig = md5(s).hexdigest()
         host = self.from_email[self.from_email.find('@') + 1:]
@@ -303,7 +323,7 @@ class TicketNotifyEmail(NotifyEmail):
         hdrs = {}
         hdrs['Message-ID'] = self.get_message_id(dest, self.modtime)
         hdrs['X-Trac-Ticket-ID'] = str(self.ticket.id)
-        hdrs['X-Trac-Ticket-URL'] = self.ticket['link']
+        hdrs['X-Trac-Ticket-URL'] = self.data['ticket']['link']
         if not self.newticket:
             msgid = self.get_message_id(dest)
             hdrs['In-Reply-To'] = msgid
