@@ -18,7 +18,6 @@
 from StringIO import StringIO
 from datetime import datetime
 import re
-from time import localtime, strftime, time
 
 from genshi.builder import tag
 
@@ -30,18 +29,17 @@ from trac.mimeview import Context
 from trac.perm import IPermissionRequestor
 from trac.resource import *
 from trac.search import ISearchSource, search_to_sql, shorten_result
-from trac.util.compat import set, sorted
-from trac.util.datefmt import parse_date, utc, to_timestamp, to_datetime, \
+from trac.util.datefmt import parse_date, utc, to_utimestamp, \
                               get_date_format_hint, get_datetime_format_hint, \
-                              format_date, format_datetime
-from trac.util.text import shorten_line, CRLF, to_unicode
-from trac.util.translation import _
+                              format_date, format_datetime, from_utimestamp
+from trac.util.text import CRLF
+from trac.util.translation import _, tag_
 from trac.ticket import Milestone, Ticket, TicketSystem, group_milestones
-from trac.ticket.query import Query, QueryModule
+from trac.ticket.query import QueryModule
 from trac.timeline.api import ITimelineEventProvider
 from trac.web import IRequestHandler, RequestDone
-from trac.web.chrome import add_link, add_notice, add_stylesheet, \
-                            add_warning, INavigationContributor
+from trac.web.chrome import add_link, add_notice, add_script, add_stylesheet, \
+                            add_warning, Chrome, INavigationContributor
 from trac.wiki.api import IWikiSyntaxProvider
 from trac.wiki.formatter import format_to
 
@@ -61,7 +59,7 @@ class TicketGroupStats(object):
         
         `title` is the display name of this group of stats (e.g.
           'ticket status').
-        `unit` is the display name of the units for these stats (e.g. 'hour').
+        `unit` is the units for these stats in plural form, e.g. _('hours')
         """
         self.title = title
         self.unit = unit
@@ -195,7 +193,8 @@ class DefaultTicketGroupStatsProvider(Component):
         for s in all_statuses:
             status_cnt[s] = 0
         if total_cnt:
-            cursor = self.env.get_db_cnx().cursor()
+            db = self.env.get_db_cnx()
+            cursor = db.cursor()
             str_ids = [str(x) for x in sorted(ticket_ids)]
             cursor.execute("SELECT status, count(status) FROM ticket "
                            "WHERE id IN (%s) GROUP BY status" %
@@ -203,7 +202,7 @@ class DefaultTicketGroupStatsProvider(Component):
             for s, cnt in cursor:
                 status_cnt[s] = cnt
 
-        stat = TicketGroupStats('ticket status', 'ticket')
+        stat = TicketGroupStats(_('ticket status'), _('tickets'))
         remaining_statuses = set(all_statuses)
         groups =  self._get_ticket_groups()
         catch_all_group = None
@@ -327,14 +326,20 @@ class RoadmapModule(Component):
         return req.path_info == '/roadmap'
 
     def process_request(self, req):
-        milestone_realm = Resource('milestone')
         req.perm.require('MILESTONE_VIEW')
 
-        showall = req.args.get('show') == 'all'
+        show = req.args.getlist('show')
+        if 'all' in show:
+            show = ['completed']
 
         db = self.env.get_db_cnx()
-        milestones = [m for m in Milestone.select(self.env, showall, db)
+        milestones = Milestone.select(self.env, 'completed' in show, db)
+        if 'noduedate' in show:
+            milestones = [m for m in milestones
+                          if m.due is not None or m.completed]
+        milestones = [m for m in milestones
                       if 'MILESTONE_VIEW' in req.perm(m.resource)]
+
         stats = []
         queries = []
 
@@ -355,8 +360,7 @@ class RoadmapModule(Component):
         username = None
         if req.authname and req.authname != 'anonymous':
             username = req.authname
-        icshref = req.href.roadmap(show=req.args.get('show'), user=username,
-                                   format='ics')
+        icshref = req.href.roadmap(show=show, user=username, format='ics')
         add_link(req, 'alternate', icshref, _('iCalendar'), 'text/calendar',
                  'ics')
 
@@ -364,8 +368,9 @@ class RoadmapModule(Component):
             'milestones': milestones,
             'milestone_stats': stats,
             'queries': queries,
-            'showall': showall,
+            'show': show,
         }
+        add_stylesheet(req, 'common/css/roadmap.css')
         return 'roadmap.html', data, None
 
     # Internal methods
@@ -391,7 +396,8 @@ class RoadmapModule(Component):
             elif status == 'assigned' or status == 'reopened':
                 return 'IN-PROCESS'
             elif status == 'closed':
-                if ticket['resolution'] == 'fixed': return 'COMPLETED'
+                if ticket['resolution'] == 'fixed':
+                    return 'COMPLETED'
                 else: return 'CANCELLED'
             else: return ''
 
@@ -404,7 +410,8 @@ class RoadmapModule(Component):
                  + ':' + escape_value(value)
             firstline = 1
             while text:
-                if not firstline: text = ' ' + text
+                if not firstline:
+                    text = ' ' + text
                 else: firstline = 0
                 buf.write(text[:75] + CRLF)
                 text = text[75:]
@@ -426,7 +433,7 @@ class RoadmapModule(Component):
                    % __version__)
         write_prop('METHOD', 'PUBLISH')
         write_prop('X-WR-CALNAME',
-                   self.config.get('project', 'name') + ' - ' + _('Roadmap'))
+                   self.env.project_name + ' - ' + _('Roadmap'))
         for milestone in milestones:
             uid = '<%s/milestone/%s@%s>' % (req.base_path, milestone.name,
                                             host)
@@ -435,9 +442,8 @@ class RoadmapModule(Component):
                 write_prop('UID', uid)
                 write_utctime('DTSTAMP', milestone.due)
                 write_date('DTSTART', milestone.due)
-                write_prop('SUMMARY', _('Milestone %(name)s') % {
-                    'name': milestone.name
-                })
+                write_prop('SUMMARY', _('Milestone %(name)s',
+                                        name=milestone.name))
                 write_prop('URL', req.base_url + '/milestone/' +
                            milestone.name)
                 if milestone.description:
@@ -455,9 +461,9 @@ class RoadmapModule(Component):
                 if milestone.due:
                     write_prop('RELATED-TO', uid)
                     write_date('DUE', milestone.due)
-                write_prop('SUMMARY', _('Ticket #%(num)s: %(summary)s') % {
-                    'num': ticket.id, 'summary': ticket['summary']
-                })
+                write_prop('SUMMARY', _('Ticket #%(num)s: %(summary)s',
+                                        num=ticket.id,
+                                        summary=ticket['summary']))
                 write_prop('URL', req.abs_href.ticket(ticket.id))
                 write_prop('DESCRIPTION', ticket['description'])
                 priority = get_priority(ticket)
@@ -472,7 +478,7 @@ class RoadmapModule(Component):
                                    (ticket.id,))
                     row = cursor.fetchone()
                     if row:
-                        write_utctime('COMPLETED', to_datetime(row[0], utc))
+                        write_utctime('COMPLETED', from_utimestamp(row[0]))
                 write_prop('END', 'VTODO')
         write_prop('END', 'VCALENDAR')
 
@@ -483,7 +489,6 @@ class RoadmapModule(Component):
         raise RequestDone
 
 
-
 class MilestoneModule(Component):
 
     implements(INavigationContributor, IPermissionRequestor, IRequestHandler,
@@ -517,7 +522,7 @@ class MilestoneModule(Component):
 
     def get_timeline_filters(self, req):
         if 'MILESTONE_VIEW' in req.perm:
-            yield ('milestone', _('Milestones'))
+            yield ('milestone', _('Milestones reached'))
 
     def get_timeline_events(self, req, start, stop, filters):
         if 'milestone' in filters:
@@ -527,11 +532,11 @@ class MilestoneModule(Component):
             # TODO: creation and (later) modifications should also be reported
             cursor.execute("SELECT completed,name,description FROM milestone "
                            "WHERE completed>=%s AND completed<=%s",
-                           (to_timestamp(start), to_timestamp(stop)))
+                           (to_utimestamp(start), to_utimestamp(stop)))
             for completed, name, description in cursor:
                 milestone = milestone_realm(id=name)
                 if 'MILESTONE_VIEW' in req.perm(milestone):
-                    yield('milestone', datetime.fromtimestamp(completed, utc),
+                    yield('milestone', from_utimestamp(completed),
                           '', (milestone, description)) # FIXME: author?
 
             # Attachments
@@ -544,7 +549,8 @@ class MilestoneModule(Component):
         if field == 'url':
             return context.href.milestone(milestone.id)
         elif field == 'title':
-            return tag('Milestone ', tag.em(milestone.id), ' completed')
+            return tag_('Milestone %(name)s completed',
+                        name=tag.em(milestone.id))
         elif field == 'description':
             return format_to(self.env, None, context(resource=milestone),
                              description)
@@ -565,8 +571,15 @@ class MilestoneModule(Component):
         add_link(req, 'up', req.href.roadmap(), _('Roadmap'))
 
         db = self.env.get_db_cnx() # TODO: db can be removed
-        milestone = Milestone(self.env, milestone_id, db)
         action = req.args.get('action', 'view')
+        try:
+            milestone = Milestone(self.env, milestone_id, db)
+        except ResourceNotFound:
+            if 'MILESTONE_CREATE' not in req.perm('milestone', milestone_id):
+                raise
+            milestone = Milestone(self.env, None, db)
+            milestone.name = milestone_id
+            action = 'edit' # rather than 'new' so that it works for POST/save
 
         if req.method == 'POST':
             if req.args.has_key('cancel'):
@@ -577,7 +590,7 @@ class MilestoneModule(Component):
             elif action == 'edit':
                 return self._do_save(req, db, milestone)
             elif action == 'delete':
-                self._do_delete(req, db, milestone)
+                self._do_delete(req, milestone)
         elif action in ('new', 'edit'):
             return self._render_editor(req, db, milestone)
         elif action == 'delete':
@@ -590,14 +603,13 @@ class MilestoneModule(Component):
 
     # Internal methods
 
-    def _do_delete(self, req, db, milestone):
+    def _do_delete(self, req, milestone):
         req.perm(milestone.resource).require('MILESTONE_DELETE')
 
         retarget_to = None
         if req.args.has_key('retarget'):
             retarget_to = req.args.get('target') or None
         milestone.delete(retarget_to, req.authname)
-        db.commit()
         add_notice(req, _('The milestone "%(name)s" has been deleted.',
                           name=milestone.name))
         req.redirect(req.href.roadmap())
@@ -611,7 +623,6 @@ class MilestoneModule(Component):
         old_name = milestone.name
         new_name = req.args.get('name')
         
-        milestone.name = new_name
         milestone.description = req.args.get('description', '')
 
         due = req.args.get('duedate', '')
@@ -628,20 +639,22 @@ class MilestoneModule(Component):
             warnings.append(msg)
 
         # -- check the name
-        if new_name:
-            if new_name != old_name:
-                # check that the milestone doesn't already exists
-                # FIXME: the whole .exists business needs to be clarified
-                #        (#4130) and should behave like a WikiPage does in
-                #        this respect.
-                try:
-                    other_milestone = Milestone(self.env, new_name, db)
-                    warn(_('Milestone "%(name)s" already exists, please '
-                           'choose another name', name=new_name))
-                except ResourceNotFound:
-                    pass
-        else:
-            warn(_('You must provide a name for the milestone.'))
+        # If the name has changed, check that the milestone doesn't already
+        # exist
+        # FIXME: the whole .exists business needs to be clarified
+        #        (#4130) and should behave like a WikiPage does in
+        #        this respect.
+        try:
+            new_milestone = Milestone(self.env, new_name, db)
+            if new_milestone.name == old_name:
+                pass        # Creation or no name change
+            elif new_milestone.name:
+                warn(_('Milestone "%(name)s" already exists, please '
+                       'choose another name.', name=new_milestone.name))
+            else:
+                warn(_('You must provide a name for the milestone.'))
+        except ResourceNotFound:
+            milestone.name = new_name
 
         # -- check completed date
         if 'completed' in req.args:
@@ -660,15 +673,16 @@ class MilestoneModule(Component):
             milestone.update()
             # eventually retarget opened tickets associated with the milestone
             if 'retarget' in req.args and completed:
-                cursor = db.cursor()
-                cursor.execute("UPDATE ticket SET milestone=%s WHERE "
-                               "milestone=%s and status != 'closed'",
-                                (retarget_to, old_name))
+                @self.env.with_transaction()
+                def retarget(db):
+                    cursor = db.cursor()
+                    cursor.execute("UPDATE ticket SET milestone=%s WHERE "
+                                   "milestone=%s and status != 'closed'",
+                                   (retarget_to, old_name))
                 self.env.log.info('Tickets associated with milestone %s '
                                   'retargeted to %s' % (old_name, retarget_to))
         else:
             milestone.insert()
-        db.commit()
 
         add_notice(req, _('Your changes have been saved.'))
         req.redirect(req.href.milestone(milestone.name))
@@ -704,6 +718,7 @@ class MilestoneModule(Component):
         else:
             req.perm(milestone.resource).require('MILESTONE_CREATE')
 
+        Chrome(self.env).add_wiki_toolbars(req)
         return 'milestone_edit.html', data, None
 
     def _render_view(self, req, db, milestone):
@@ -782,6 +797,8 @@ class MilestoneModule(Component):
                     percent = float(gstat.count) / float(max_count) * 100
                 gs_dict['percent_of_max_total'] = percent
 
+        add_stylesheet(req, 'common/css/roadmap.css')
+        add_script(req, 'common/js/folding.js')
         return 'milestone_view.html', data, None
 
     # IWikiSyntaxProvider methods
@@ -831,6 +848,26 @@ class MilestoneModule(Component):
         else:
             return desc
 
+    def resource_exists(self, resource):
+        """
+        >>> from trac.test import EnvironmentStub
+        >>> env = EnvironmentStub()
+        
+        >>> m1 = Milestone(env)
+        >>> m1.name = 'M1'
+        >>> m1.insert()
+        
+        >>> MilestoneModule(env).resource_exists(Resource('milestone', 'M1'))
+        True
+        >>> MilestoneModule(env).resource_exists(Resource('milestone', 'M2'))
+        False
+        """
+        db = self.env.get_read_db()
+        cursor = db.cursor()
+        cursor.execute("SELECT name FROM milestone WHERE name=%s",
+                       (resource.id,))
+        return bool(cursor.fetchall())
+
     # ISearchSource methods
 
     def get_search_filters(self, req):
@@ -851,10 +888,10 @@ class MilestoneModule(Component):
         for name, due, completed, description in cursor:
             milestone = milestone_realm(id=name)
             if 'MILESTONE_VIEW' in req.perm(milestone):
+                dt = (completed and from_utimestamp(completed) or
+                      due and from_utimestamp(due) or datetime.now(utc))
                 yield (get_resource_url(self.env, milestone, req.href),
-                       get_resource_name(self.env, milestone),
-                       datetime.fromtimestamp(
-                           completed or due or time(), utc),
+                       get_resource_name(self.env, milestone), dt,
                        '', shorten_result(description, terms))
         
         # Attachments
