@@ -18,26 +18,21 @@
 #         Matthew Good <trac@matt-good.net>
 
 import errno
+import inspect
+from itertools import izip, tee
 import locale
 import os.path
+from pkg_resources import find_distributions
+import random
 import re
+import shutil
 import sys
-import time
 import tempfile
+import time
 from urllib import quote, unquote, urlencode
-from itertools import izip
 
-# Imports for backward compatibility
-from trac.core import TracError
-from trac.util.compat import md5, reversed, sha1, sorted, tee
-from trac.util.html import escape, unescape, Markup, Deuglifier
-from trac.util.text import CRLF, to_utf8, to_unicode, shorten_line, \
-                           wrap, pretty_size
-from trac.util.datefmt import pretty_timedelta, format_datetime, \
-                              format_date, format_time, \
-                              get_date_format_hint, \
-                              get_datetime_format_hint, http_date, \
-                              parse_date
+from trac.util.compat import any, md5, sha1, sorted
+from trac.util.text import to_unicode
 
 # -- req/session utils
 
@@ -86,54 +81,62 @@ except NameError:
 
 can_rename_open_file = False
 if os.name == 'nt':
+    _rename = lambda src, dst: False
+    _rename_atomic = lambda src, dst: False
+    
     try:
         import ctypes
         MOVEFILE_REPLACE_EXISTING = 0x1
         MOVEFILE_WRITE_THROUGH = 0x8
+        MoveFileEx = ctypes.windll.kernel32.MoveFileExW
         
-        try:
-            MoveFileTransacted = ctypes.windll.kernel32.MoveFileTransactedA
-            CreateTransaction = ctypes.windll.ktmw32.CreateTransaction
-            CommitTransaction = ctypes.windll.ktmw32.CommitTransaction
-            CloseHandle = ctypes.windll.kernel32.CloseHandle
-            can_rename_open_file = True
-            
-            def rename(src, dst):
-                ta = CreateTransaction(None, 0, 0, 0, 0, 1000, 
-                                       'Trac forced rename')
-                if ta == -1:
-                    raise ctypes.WinError()
-                try:
-                    if not (MoveFileTransacted(src, dst, None, None,
-                                               MOVEFILE_REPLACE_EXISTING
-                                               | MOVEFILE_WRITE_THROUGH, ta) \
-                            and CommitTransaction(ta)):
-                        raise ctypes.WinError()
-                finally:
-                    CloseHandle(ta)
-        except AttributeError:
-            MoveFileEx = ctypes.windll.kernel32.MoveFileExA
-            
-            def rename(src, dst):
-                if not MoveFileEx(src, dst, MOVEFILE_REPLACE_EXISTING
-                                            | MOVEFILE_WRITE_THROUGH):
-                    raise ctypes.WinError()
-    except Exception:
-        import random
+        def _rename(src, dst):
+            if not isinstance(src, unicode):
+                src = unicode(src, sys.getfilesystemencoding())
+            if not isinstance(dst, unicode):
+                dst = unicode(dst, sys.getfilesystemencoding())
+            if _rename_atomic(src, dst):
+                return True
+            return MoveFileEx(src, dst, MOVEFILE_REPLACE_EXISTING
+                                        | MOVEFILE_WRITE_THROUGH)
         
-        def rename(src, dst):
+        CreateTransaction = ctypes.windll.ktmw32.CreateTransaction
+        CommitTransaction = ctypes.windll.ktmw32.CommitTransaction
+        MoveFileTransacted = ctypes.windll.kernel32.MoveFileTransactedW
+        CloseHandle = ctypes.windll.kernel32.CloseHandle
+        can_rename_open_file = True
+        
+        def _rename_atomic(src, dst):
+            ta = CreateTransaction(None, 0, 0, 0, 0, 10000, 'Trac rename')
+            if ta == -1:
+                return False
             try:
-                os.rename(src, dst)
-            except OSError, e:
-                if e.errno != errno.EEXIST:
-                    raise
-                old = "%s-%08x" % (dst, random.randint(0, sys.maxint))
-                os.rename(dst, old)
-                os.rename(src, dst)
-                try:
-                    os.unlink(old)
-                except Exception:
-                    pass
+                return (MoveFileTransacted(src, dst, None, None,
+                                           MOVEFILE_REPLACE_EXISTING
+                                           | MOVEFILE_WRITE_THROUGH, ta)
+                        and CommitTransaction(ta))
+            finally:
+                CloseHandle(ta)
+    except Exception:
+        pass
+    
+    def rename(src, dst):
+        # Try atomic or pseudo-atomic rename
+        if _rename(src, dst):
+            return
+        # Fall back to "move away and replace"
+        try:
+            os.rename(src, dst)
+        except OSError, e:
+            if e.errno != errno.EEXIST:
+                raise
+            old = "%s-%08x" % (dst, random.randint(0, sys.maxint))
+            os.rename(dst, old)
+            os.rename(src, dst)
+            try:
+                os.unlink(old)
+            except Exception:
+                pass
 else:
     rename = os.rename
     can_rename_open_file = True
@@ -148,6 +151,7 @@ class AtomicFile(object):
     the temporary file is removed.
     """
     def __init__(self, path, mode='w', bufsize=-1):
+        self._file = None
         self._path = path
         (dir, name) = os.path.split(path)
         (fd, self._temp) = tempfile.mkstemp(prefix=name + '-', dir=dir)
@@ -276,10 +280,75 @@ class NaivePopen:
             if capturestderr and os.path.isfile(errfile):
                 os.remove(errfile)
 
+
+def makedirs(path, overwrite=False):
+    if overwrite and os.path.exists(path):
+        return
+    os.makedirs(path)
+
+
+def copytree(src, dst, symlinks=False, skip=[], overwrite=False):
+    """Recursively copy a directory tree using copy2() (from shutil.copytree.)
+
+    Added a `skip` parameter consisting of absolute paths
+    which we don't want to copy.
+    """
+    def str_path(path):
+        if isinstance(path, unicode):
+            path = path.encode(sys.getfilesystemencoding() or
+                               locale.getpreferredencoding())
+        return path
+
+    def remove_if_overwriting(path):
+        if overwrite and os.path.exists(path):
+            os.unlink(path)
+
+    skip = [str_path(f) for f in skip]
+    def copytree_rec(src, dst):
+        names = os.listdir(src)
+        makedirs(dst, overwrite=overwrite)
+        errors = []
+        for name in names:
+            srcname = os.path.join(src, name)
+            if srcname in skip:
+                continue
+            dstname = os.path.join(dst, name)
+            try:
+                if symlinks and os.path.islink(srcname):
+                    remove_if_overwriting(dstname)
+                    linkto = os.readlink(srcname)
+                    os.symlink(linkto, dstname)
+                elif os.path.isdir(srcname):
+                    copytree_rec(srcname, dstname)
+                else:
+                    remove_if_overwriting(dstname)
+                    shutil.copy2(srcname, dstname)
+                # XXX What about devices, sockets etc.?
+            except (IOError, OSError), why:
+                errors.append((srcname, dstname, str(why)))
+            # catch the Error from the recursive copytree so that we can
+            # continue with other files
+            except shutil.Error, err:
+                errors.extend(err.args[0])
+        try:
+            shutil.copystat(src, dst)
+        except WindowsError, why:
+            pass # Ignore errors due to limited Windows copystat support
+        except OSError, why:
+            errors.append((src, dst, str(why)))
+        if errors:
+            raise shutil.Error(errors)
+    copytree_rec(str_path(src), str_path(dst))
+
+
 # -- sys utils
 
 def arity(f):
-    return f.func_code.co_argcount
+    """Return the number of arguments expected by the given function, unbound
+    or bound method.
+    """
+    return f.func_code.co_argcount - bool(getattr(f, 'im_self', False))
+
 
 def get_last_traceback():
     import traceback
@@ -288,38 +357,85 @@ def get_last_traceback():
     traceback.print_exc(file=tb)
     return to_unicode(tb.getvalue())
 
-def get_lines_from_file(filename, lineno, context=0):
+
+_egg_path_re = re.compile(r'build/bdist\.[^/]+/egg/(.*)')
+
+def get_lines_from_file(filename, lineno, context=0, globals=None):
     """Return `content` number of lines before and after the specified
-    `lineno` from the file identified by `filename`.
+    `lineno` from the (source code) file identified by `filename`.
     
     Returns a `(lines_before, line, lines_after)` tuple.
     """
-    if os.path.isfile(filename):
-        fileobj = open(filename, 'U')
-        try:
-            lines = fileobj.readlines()
-            lbound = max(0, lineno - context)
-            ubound = lineno + 1 + context
-
-
-            charset = None
-            rep = re.compile('coding[=:]\s*([-\w.]+)')
-            for linestr in lines[0], lines[1]:
-                match = rep.search(linestr)
-                if match:
-                    charset = match.group(1)
+    # The linecache module can load source code from eggs since Python 2.6.
+    # Prior versions return lines from the wrong file, so we try locating
+    # the file in eggs manually first.
+    lines = []
+    match = _egg_path_re.match(filename)
+    if match:
+        import zipfile
+        for path in sys.path:
+            try:
+                zip = zipfile.ZipFile(path, 'r')
+                try:
+                    lines = zip.read(match.group(1)).splitlines()
                     break
+                finally:
+                    zip.close()
+            except Exception:
+                pass
 
-            before = [to_unicode(l.rstrip('\n'), charset)
-                         for l in lines[lbound:lineno]]
-            line = to_unicode(lines[lineno].rstrip('\n'), charset)
-            after = [to_unicode(l.rstrip('\n'), charset) \
-                         for l in lines[lineno + 1:ubound]]
+    if not lines:
+        import linecache
+        linecache.checkcache(filename)
+        if arity(linecache.getlines) >= 2:
+            lines = linecache.getlines(filename, globals)
+        else:   # Python 2.4
+            lines = linecache.getlines(filename)
+
+    if not 0 <= lineno < len(lines):
+        return (), None, ()
+    lbound = max(0, lineno - context)
+    ubound = lineno + 1 + context
+
+    charset = None
+    rep = re.compile('coding[=:]\s*([-\w.]+)')
+    for linestr in lines[:2]:
+        match = rep.search(linestr)
+        if match:
+            charset = match.group(1)
+            break
+
+    before = [to_unicode(l.rstrip('\n'), charset)
+                 for l in lines[lbound:lineno]]
+    line = to_unicode(lines[lineno].rstrip('\n'), charset)
+    after = [to_unicode(l.rstrip('\n'), charset) \
+                 for l in lines[lineno + 1:ubound]]
+
+    return before, line, after
+
+
+def get_frame_info(tb):
+    """Return frame information for a traceback."""
+    frames = []
+    while tb:
+        tb_hide = tb.tb_frame.f_locals.get('__traceback_hide__')
+        if tb_hide in ('before', 'before_and_this'):
+            del frames[:]
+            tb_hide = tb_hide[6:]
+        if not tb_hide:
+            filename = tb.tb_frame.f_code.co_filename
+            filename = filename.replace('\\', '/')
+            lineno = tb.tb_lineno - 1
+            before, line, after = get_lines_from_file(filename, lineno, 5,
+                                                      tb.tb_frame.f_globals)
+            frames.append({'traceback': tb, 'filename': filename,
+                           'lineno': lineno, 'line': line,
+                           'lines_before': before, 'lines_after': after,
+                           'function': tb.tb_frame.f_code.co_name,
+                           'vars': tb.tb_frame.f_locals})
+        tb = tb.tb_next
+    return frames
 
-            return before, line, after
-        finally:
-            fileobj.close()
-    return (), None, ()
 
 def safe__import__(module_name):
     """
@@ -339,10 +455,24 @@ def safe__import__(module_name):
                 del(sys.modules[modname])
         raise e
 
+
+def get_doc(obj):
+    """Return the docstring of an object as a tuple `(summary, description)`,
+    where `summary` is the first paragraph and `description` is the remaining
+    text.
+    """
+    doc = inspect.getdoc(obj)
+    if not doc:
+        return (None, None)
+    doc = to_unicode(doc).split('\n\n', 1)
+    summary = doc[0].replace('\n', ' ')
+    description = len(doc) > 1 and doc[1] or None
+    return (summary, description)
+
 # -- setuptools utils
 
 def get_module_path(module):
-    # Determine the plugin that this component belongs to
+    """Return the base path the given module is imported from"""
     path = module.__file__
     module_name = module.__name__
     if path.endswith('.pyc') or path.endswith('.pyo'):
@@ -357,6 +487,23 @@ def get_module_path(module):
             break
     return base_path
 
+def get_sources(path):
+    """Return a dictionary mapping Python module source paths to the
+    distributions that contain them.
+    """
+    sources = {}
+    for dist in find_distributions(path, only=True):
+        try:
+            toplevels = dist.get_metadata('top_level.txt').splitlines()
+            toplevels = [each + '/' for each in toplevels]
+            files = dist.get_metadata('SOURCES.txt').splitlines()
+            sources.update((src, dist) for src in files
+                           if any(src.startswith(toplevel)
+                                  for toplevel in toplevels))
+        except (KeyError, IOError):
+            pass    # Metadata not found
+    return sources
+
 def get_pkginfo(dist):
     """Get a dictionary containing package information for a package
 
@@ -369,17 +516,13 @@ def get_pkginfo(dist):
     """
     import types
     if isinstance(dist, types.ModuleType):
-        try:
-            from pkg_resources import find_distributions
-            module = dist
-            module_path = get_module_path(module)
-            for dist in find_distributions(module_path, only=True):
-                if os.path.isfile(module_path) or \
-                       dist.key == module.__name__.lower():
-                    break
-            else:
-                return {}
-        except ImportError:
+        module = dist
+        module_path = get_module_path(module)
+        for dist in find_distributions(module_path, only=True):
+            if os.path.isfile(module_path) or \
+                   dist.key == module.__name__.lower():
+                break
+        else:
             return {}
     import email
     attrs = ('author', 'author-email', 'license', 'home-page', 'summary',
@@ -404,7 +547,6 @@ def get_pkginfo(dist):
 # -- crypto utils
 
 def hex_entropy(bytes=32):
-    import random
     return sha1(str(random.random())).hexdigest()[:bytes]
 
 
@@ -467,11 +609,13 @@ def md5crypt(password, salt, magic='$1$'):
     for a, b, c in ((0, 6, 12), (1, 7, 13), (2, 8, 14), (3, 9, 15), (4, 10, 5)):
         v = ord(final[a]) << 16 | ord(final[b]) << 8 | ord(final[c])
         for i in range(4):
-            rearranged += itoa64[v & 0x3f]; v >>= 6
+            rearranged += itoa64[v & 0x3f]
+            v >>= 6
 
     v = ord(final[11])
     for i in range(2):
-        rearranged += itoa64[v & 0x3f]; v >>= 6
+        rearranged += itoa64[v & 0x3f]
+        v >>= 6
 
     return magic + salt + '$' + rearranged
 
@@ -706,7 +850,9 @@ def content_disposition(type, filename=None):
     return type
 
 def pairwise(iterable):
-    """s -> (s0,s1), (s1,s2), (s2, s3), ...
+    """
+    >>> list(pairwise([0, 1, 2, 3]))
+    [(0, 1), (1, 2), (2, 3)]
 
     :deprecated: since 0.11 (if this really needs to be used, rewrite it
                              without izip)
@@ -719,6 +865,12 @@ def pairwise(iterable):
     return izip(a, b)
 
 def partition(iterable, order=None):
+    """
+    >>> partition([(1,"a"),(2, "b"),(3, "a")])
+    {'a': [1, 3], 'b': [2]}
+    >>> partition([(1,"a"),(2, "b"),(3, "a")], "ab")
+    [[1, 3], [2]]
+    """
     result = {}
     if order is not None:
         for key in order:
@@ -728,3 +880,40 @@ def partition(iterable, order=None):
     if order is None:
         return result
     return [result[key] for key in order]
+
+def as_int(s, default, min=None, max=None):
+    """Convert s to an int and limit it to the given range, or return default
+    if unsuccessful."""
+    try:
+        value = int(s)
+    except (TypeError, ValueError):
+        return default
+    if min is not None and value < min:
+        value = min
+    if max is not None and value > max:
+        value = max
+    return value
+
+def as_bool(s):
+    """Convert s to a `bool`."""
+    try:
+        return bool(int(s))
+    except (TypeError, ValueError):
+        return s.lower() in ('true', 'yes', 'on')
+
+def pathjoin(*args):
+    """Strip `/` from the arguments and join them with a single `/`."""
+    return '/'.join(filter(None, (each.strip('/') for each in args if each)))
+
+
+# Imports for backward compatibility (at bottom to avoid circular dependencies)
+from trac.core import TracError
+from trac.util.compat import reversed
+from trac.util.html import escape, unescape, Markup, Deuglifier
+from trac.util.text import CRLF, to_utf8, shorten_line, wrap, pretty_size
+from trac.util.datefmt import pretty_timedelta, format_datetime, \
+                              format_date, format_time, \
+                              get_date_format_hint, \
+                              get_datetime_format_hint, http_date, \
+                              parse_date
+
