@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2005 Edgewall Software
+# Copyright (C) 2003-2009 Edgewall Software
 # Copyright (C) 2003-2004 Jonas Borgström <jonas@edgewall.com>
 # Copyright (C) 2005 Christopher Lenz <cmlenz@gmx.de>
 # All rights reserved.
@@ -16,13 +16,11 @@
 # Author: Jonas Borgström <jonas@edgewall.com>
 #         Christopher Lenz <cmlenz@gmx.de>
 
-from urllib import quote, urlencode
 from trac.util.text import unicode_quote, unicode_urlencode
 
 
 class Href(object):
-    """
-    Implements a callable that constructs URLs with the given base. The
+    """Implements a callable that constructs URLs with the given base. The
     function can be called with any number of positional and keyword
     arguments which than are used to assemble the URL.
 
@@ -76,6 +74,18 @@ class Href(object):
 
     >>> href('timeline', {'from': '02/24/05', 'daysback': 30})
     '/trac/timeline?daysback=30&from=02%2F24%2F05'
+    >>> href('timeline', {})
+    '/trac/timeline'
+    >>> href('timeline', [('from', '02/24/05')])
+    '/trac/timeline?from=02%2F24%2F05'
+    >>> href('timeline', ()) == href('timeline', []) == href('timeline', {})
+    True
+
+    The usual way of quoting arguments that would otherwise be interpreted
+    as Python keywords is supported too:
+
+    >>> href('timeline', from_='02/24/05', daysback=30)
+    '/trac/timeline?from=02%2F24%2F05&daysback=30'
 
     If the order of query string parameters should be preserved, you may also
     pass a sequence of (name, value) tuples as last positional argument:
@@ -111,42 +121,42 @@ class Href(object):
     """
 
     def __init__(self, base):
-        self.base = base
+        self.base = base.rstrip('/')
         self._derived = {}
 
     def __call__(self, *args, **kw):
         href = self.base
-        if href and href[-1] == '/':
-            href = href[:-1]
         params = []
 
         def add_param(name, value):
-            if type(value) in (list, tuple):
-                for i in [i for i in value if i != None]:
+            if isinstance(value, (list, tuple)):
+                for i in [i for i in value if i is not None]:
                     params.append((name, i))
-            elif v != None:
+            elif value is not None:
                 params.append((name, value))
 
         if args:
             lastp = args[-1]
-            if lastp and type(lastp) is dict:
-                for k,v in lastp.items():
+            if isinstance(lastp, dict):
+                for k, v in lastp.items():
                     add_param(k, v)
                 args = args[:-1]
-            elif lastp and type(lastp) in (list, tuple):
-                for k,v in lastp:
+            elif isinstance(lastp, (list, tuple)):
+                for k, v in lastp:
                     add_param(k, v)
                 args = args[:-1]
 
         # build the path
         path = '/'.join([unicode_quote(unicode(arg).strip('/')) for arg in args
-                         if arg != None])
+                         if arg is not None])
         if path:
             href += '/' + path
+        elif not href:
+            href = '/'
 
         # assemble the query string
-        for k,v in kw.items():
-            add_param(k, v)
+        for k, v in kw.items():
+            add_param(k.endswith('_') and k[:-1] or k, v)
 
         if params:
             href += '?' + unicode_urlencode(params)
@@ -154,10 +164,17 @@ class Href(object):
         return href
 
     def __getattr__(self, name):
-        if not self._derived.has_key(name):
+        if name not in self._derived:
             self._derived[name] = lambda *args, **kw: self(name, *args, **kw)
         return self._derived[name]
 
+    def __add__(self, rhs):
+        if rhs.startswith('/'):
+            return self.base + rhs
+        if rhs:
+            return self.base + '/' + rhs
+        return self.base or '/'
+
 
 if __name__ == '__main__':
     import doctest, sys
