@@ -16,22 +16,14 @@
 # Author: Jonas Borgström <jonas@edgewall.com>
 #         Christopher Lenz <cmlenz@gmx.de>
 
-try:
-    import threading
-except ImportError:
-    import dummy_threading as threading
-import time
-import urllib
 import re
-from StringIO import StringIO
 
 from genshi.builder import tag
 
+from trac.cache import cached
 from trac.config import BoolOption
 from trac.core import *
 from trac.resource import IResourceManager
-from trac.util import reversed
-from trac.util.html import html
 from trac.util.translation import _
 from trac.wiki.parser import WikiParser
 
@@ -53,6 +45,9 @@ class IWikiChangeListener(Interface):
     def wiki_page_version_deleted(page):
         """Called when a version of a page has been deleted."""
 
+    def wiki_page_renamed(page, old_name): 
+        """Called when a page has been renamed.""" 
+
 
 class IWikiPageManipulator(Interface):
     """Extension point interface for components that need to do specific
@@ -63,34 +58,69 @@ class IWikiPageManipulator(Interface):
     """
 
     def prepare_wiki_page(req, page, fields):
-        """Not currently called, but should be provided for future
-        compatibility."""
+        """Validate a wiki page before rendering it.
+
+        `page` is the `WikiPage` being viewed.
+        `fields` is a dictionary which contains the wiki `text` of the page,
+        initially identical to `page.text` but it  can eventually be
+        transformed in place before being used as input to the formatter.
+        """
 
     def validate_wiki_page(req, page):
         """Validate a wiki page after it's been populated from user input.
+
+        `page` is the `WikiPage` being edited.
         
         Must return a list of `(field, message)` tuples, one for each problem
         detected. `field` can be `None` to indicate an overall problem with the
-        page. Therefore, a return value of `[]` means everything is OK."""
+        page. Therefore, a return value of `[]` means everything is OK.
+        """
 
 
 class IWikiMacroProvider(Interface):
-    """Extension point interface for components that provide Wiki macros."""
+    """Extension point interface for components that provide Wiki macros
+
+    New Wiki processors can also be added that way. (''since 0.12'')
+    """
 
     def get_macros():
-        """Return an iterable that provides the names of the provided macros."""
+        """Return an iterable that provides the names of the provided macros.
+        """
 
     def get_macro_description(name):
-        """Return a plain text description of the macro with the specified name.
-        """
+        """Return a plain text description of the macro with the specified
+        name."""
 
     def render_macro(req, name, content):
         """Return the HTML output of the macro (deprecated)"""
 
-    def expand_macro(formatter, name, content):
+    def expand_macro(formatter, name, content, args=None):
         """Called by the formatter when rendering the parsed wiki text.
 
-        (since 0.11)
+        This form is preferred over `render_macro`, as you get the
+        formatter, which knows the current `.context` (and the `.req`,
+        but ideally you shouldn't use it in your macros). (''since 0.11'')
+
+        `name` is the name by which the macro has been called; remember
+        that via `get_macros`, multiple names could be associated to this
+        macros. Note that the macro names are case sensitive.
+        
+        `content` is the content of the macro call. When called using macro
+        syntax (`[[Macro(content)]]`), this is the string contained between
+        parentheses, usually containing macro arguments. When called using wiki
+        processor syntax (`{{{!#Macro ...}}}`), it is the content of the
+        processor block, that is, the text starting on the line following the
+        macro name.
+
+        `args` will be a dictionary containing the named parameters which you
+        can specify when calling the macro using the wiki processor syntax:
+        `{{{#!Macro arg1=value1 arg2="value 2"`. In this example, `args` will
+        be `{'arg1': 'value1', 'arg2': 'value 2'}`).
+        If no named parameters are given, `args` will be `{}`. That makes it
+        possible to differentiate with a call using the macro syntax, in which
+        case `args` will be `None` (see `parse_args` for a convenient way to
+        extract arguments and name parameters from the `content` inside the
+        parentheses, in the latter situation). (''since 0.12'')
         """
 
 
@@ -108,10 +138,14 @@ class IWikiSyntaxProvider(Interface):
     def get_link_resolvers():
         """Return an iterable over (namespace, formatter) tuples.
 
-        Each formatter should be a function of the form
-        fmt(formatter, ns, target, label), and should
-        return some HTML fragment.
-        The `label` is already HTML escaped, whereas the `target` is not.
+        Each formatter should be a function of the form:
+        {{{#!python
+        def format(formatter, ns, target, label, fullmatch=None):
+            pass
+        }}}
+        and should return some HTML fragment. The `label` is already
+        HTML escaped, whereas the `target` is not. The `fullmatch` argument
+        is optional, and is bound to the regexp match object for the link.
         """
 
 
@@ -155,18 +189,15 @@ def parse_args(args, strict=True):
     return largs, kwargs
 
 
-
 class WikiSystem(Component):
-    """Represents the wiki system."""
+    """Wiki system manager."""
 
-    implements(IWikiChangeListener, IWikiSyntaxProvider, IResourceManager)
+    implements(IWikiSyntaxProvider, IResourceManager)
 
     change_listeners = ExtensionPoint(IWikiChangeListener)
     macro_providers = ExtensionPoint(IWikiMacroProvider)
     syntax_providers = ExtensionPoint(IWikiSyntaxProvider)
 
-    INDEX_UPDATE_INTERVAL = 5 # seconds
-
     ignore_missing_pages = BoolOption('wiki', 'ignore_missing_pages', 'false',
         """Enable/disable highlighting CamelCase links to missing pages
         (''since 0.9'').""")
@@ -183,26 +214,12 @@ class WikiSystem(Component):
         For public sites where anonymous users can edit the wiki it is
         recommended to leave this option disabled (which is the default).""")
 
-    def __init__(self):
-        self._index = None
-        self._last_index_update = 0
-        self._index_lock = threading.RLock()
-
-    def _update_index(self):
-        self._index_lock.acquire()
-        try:
-            now = time.time()
-            if now > self._last_index_update + WikiSystem.INDEX_UPDATE_INTERVAL:
-                self.log.debug('Updating wiki page index')
-                db = self.env.get_db_cnx()
-                cursor = db.cursor()
-                cursor.execute("SELECT DISTINCT name FROM wiki")
-                self._index = {}
-                for (name,) in cursor:
-                    self._index[name] = True
-                self._last_index_update = now
-        finally:
-            self._index_lock.release()
+    @cached
+    def pages(self, db):
+        """Return the names of all existing wiki pages."""
+        cursor = db.cursor()
+        cursor.execute("SELECT DISTINCT name FROM wiki")
+        return set(row[0] for row in cursor)
 
     # Public API
 
@@ -212,35 +229,13 @@ class WikiSystem(Component):
         If the `prefix` parameter is given, only names that start with that
         prefix are included.
         """
-        self._update_index()
-        # Note: use of keys() is intentional since iterkeys() is prone to
-        # errors with concurrent modification
-        for page in self._index.keys():
+        for page in self.pages:
             if not prefix or page.startswith(prefix):
                 yield page
 
     def has_page(self, pagename):
         """Whether a page with the specified name exists."""
-        self._update_index()
-        return self._index.has_key(pagename.rstrip('/'))
-
-    # IWikiChangeListener methods
-
-    def wiki_page_added(self, page):
-        if not self.has_page(page.name):
-            self.log.debug('Adding page %s to index' % page.name)
-            self._index[page.name] = True
-
-    def wiki_page_changed(self, page, version, t, comment, author, ipnr):
-        pass
-
-    def wiki_page_deleted(self, page):
-        if self.has_page(page.name):
-            self.log.debug('Removing page %s from index' % page.name)
-            del self._index[page.name]
-
-    def wiki_page_version_deleted(self, page):
-        pass
+        return pagename.rstrip('/') in self.pages
 
     # IWikiSyntaxProvider methods
 
@@ -249,29 +244,44 @@ class WikiSystem(Component):
     # here adapted to exclude terminal "." and ":" characters
 
     PAGE_SPLIT_RE = re.compile(r"([a-z])([A-Z])(?=[a-z])")
-    
+
+    Lu = ''.join(unichr(c) for c in range(0, 0x10000) if unichr(c).isupper())
+    Ll = ''.join(unichr(c) for c in range(0, 0x10000) if unichr(c).islower())
+
     def format_page_name(self, page, split=False):
         if split or self.split_page_names:
             return self.PAGE_SPLIT_RE.sub(r"\1 \2", page)
         return page
 
-    def get_wiki_syntax(self):
-        from trac.wiki.formatter import Formatter
-        lower = r'(?<![A-Z0-9_])' # No Upper case when looking behind
-        upper = r'(?<![a-z0-9_])' # No Lower case when looking behind
-        wiki_page_name = (
-            r"\w%s(?:\w%s)+(?:\w%s(?:\w%s)*[\w/]%s)+" % # wiki words
-            (upper, lower, upper, lower, lower) +
-            r"(?:@\d+)?" # optional version
-            r"(?:#%s)?" % self.XML_NAME + # optional fragment id
-            r"(?=:(?:\Z|\s)|[^:a-zA-Z]|\s|\Z)" # what should follow it
-            )
+    def make_label_from_target(self, target):
+        """Create a label from a wiki target.
+        
+        A trailing fragment and query string is stripped. Then, leading `./`,
+        `../` and '/' elements are stripped, except when this would lead to an
+        empty label. Finally, if `[wiki] split_page_names` is true, the label
+        is split accordingly.
+        """
+        label = target.split('#', 1)[0].split('?', 1)[0]
+        if not label:
+            return target
+        components = label.split('/')
+        for i, comp in enumerate(components):
+            if comp not in ('', '.', '..'):
+                label = '/'.join(components[i:])
+                break
+        return self.format_page_name(label)
 
+    def get_wiki_syntax(self):
+        wiki_page_name = (
+            r"(?:[%(upper)s](?:[%(lower)s])+/?){2,}" # wiki words
+            r"(?:@\d+)?"                             # optional version
+            r"(?:#%(xml)s)?"                         # optional fragment id
+            r"(?=:(?:\Z|\s)|[^:%(upper)s%(lower)s]|\s|\Z)"
+            # what should follow it
+            % {'upper': self.Lu, 'lower': self.Ll, 'xml': self.XML_NAME})
         
         # Regular WikiPageNames
         def wikipagename_link(formatter, match, fullmatch):
-            if not _check_unicode_camelcase(match):
-                return match
             return self._format_link(formatter, 'wiki', match,
                                      self.format_page_name(match),
                                      self.ignore_missing_pages, match)
@@ -283,21 +293,32 @@ class WikiSystem(Component):
         def wikipagename_with_label_link(formatter, match, fullmatch):
             page = fullmatch.group('wiki_page')
             label = fullmatch.group('wiki_label')
-            if not _check_unicode_camelcase(page):
-                return label
             return self._format_link(formatter, 'wiki', page, label.strip(),
                                      self.ignore_missing_pages, match)
         yield (r"!?\[(?P<wiki_page>%s)\s+(?P<wiki_label>%s|[^\]]+)\]"
                % (wiki_page_name, WikiParser.QUOTED_STRING),
                wikipagename_with_label_link)
 
-        # MoinMoin's ["internal free link"] 
+        # MoinMoin's ["internal free link"] and ["free link" with label]
         def internal_free_link(fmt, m, fullmatch): 
-            return self._format_link(fmt, 'wiki', m[2:-2], m[2:-2], False) 
-        yield (r"!?\[(?:%s)\]" % WikiParser.QUOTED_STRING, internal_free_link) 
+            page = fullmatch.group('ifl_page')[1:-1]
+            label = fullmatch.group('ifl_label')
+            if label is None:
+                label = self.make_label_from_target(page)
+            return self._format_link(fmt, 'wiki', page, label.strip(), False)
+        yield (r"!?\[(?P<ifl_page>%s)(?:\s+(?P<ifl_label>%s|[^\]]+))?\]"
+               % (WikiParser.QUOTED_STRING, WikiParser.QUOTED_STRING),
+               internal_free_link)
 
     def get_link_resolvers(self):
-        def link_resolver(formatter, ns, target, label):
+        def link_resolver(formatter, ns, target, label, fullmatch=None):
+            if fullmatch is not None:
+                # If no explicit label was specified for a [wiki:...] link,
+                # generate a "nice" label instead of keeping the label
+                # generated by the Formatter (usually the target itself).
+                groups = fullmatch.groupdict()
+                if groups.get('lns') and not groups.get('label'):
+                    label = self.make_label_from_target(target)
             return self._format_link(formatter, ns, target, label, False)
         yield ('wiki', link_resolver)
 
@@ -309,7 +330,17 @@ class WikiSystem(Component):
             pagename, version = pagename.split('@', 1)
         if version and query:
             query = '&' + query[1:]
-        pagename = pagename.strip('/') or 'WikiStart'
+        pagename = pagename.rstrip('/') or 'WikiStart'
+        referrer = ''
+        if formatter.resource and formatter.resource.realm == 'wiki':
+            referrer = formatter.resource.id
+        if pagename.startswith('/'):
+            pagename = pagename.lstrip('/')
+        elif pagename.startswith('./') or pagename.startswith('../') \
+                                                or pagename in ('.', '..'):
+            pagename = self._resolve_relative_name(pagename, referrer)
+        else:
+            pagename = self._resolve_scoped_name(pagename, referrer)
         if 'WIKI_VIEW' in formatter.perm('wiki', pagename, version):
             href = formatter.href.wiki(pagename, version=version) + query \
                    + fragment
@@ -324,11 +355,47 @@ class WikiSystem(Component):
                 else:
                     return tag.a(label + '?', class_='missing wiki')
         elif ignore_missing and not self.has_page(pagename):
-            return label
+            return original_label or label
         else:
             return tag.a(label, class_='forbidden wiki',
                          title=_("no permission to view this wiki page"))
 
+    def _resolve_relative_name(self, pagename, referrer):
+        base = referrer.split('/')
+        components = pagename.split('/')
+        for i, comp in enumerate(components):
+            if comp == '..':
+                if base:
+                    base.pop()
+            elif comp and comp != '.':
+                base.extend(components[i:])
+                break
+        return '/'.join(base)
+    
+    def _resolve_scoped_name(self, pagename, referrer):
+        referrer = referrer.split('/')
+        if len(referrer) == 1:           # Non-hierarchical referrer
+            return pagename
+        # Test for pages with same name, higher in the hierarchy
+        for i in range(len(referrer) - 1, 0, -1):
+            name = '/'.join(referrer[:i]) + '/' + pagename
+            if self.has_page(name):
+                return name
+        if self.has_page(pagename):
+            return pagename
+        # If we are on First/Second/Third, and pagename is Second/Other,
+        # resolve to First/Second/Other instead of First/Second/Second/Other
+        # See http://trac.edgewall.org/ticket/4507#comment:12
+        if '/' in pagename:
+            (first, rest) = pagename.split('/', 1)
+            for (i, part) in enumerate(referrer):
+                if first == part:
+                    anchor = '/'.join(referrer[:i + 1])
+                    if self.has_page(anchor):
+                        return anchor + '/' + rest
+        # Assume the user wants a sibling of referrer
+        return '/'.join(referrer[:-1]) + '/' + pagename
+
     # IResourceManager methods
 
     def get_resource_realms(self):
@@ -355,30 +422,26 @@ class WikiSystem(Component):
         """
         return self.format_page_name(resource.id)
 
+    def resource_exists(self, resource):
+        """
+        >>> from trac.test import EnvironmentStub
+        >>> from trac.resource import Resource, resource_exists
+        >>> env = EnvironmentStub()
 
-def _check_unicode_camelcase(pagename):
-    """A camelcase word must have at least 2 humps (well...)
-
-    >>> _check_unicode_camelcase(u"\xc9l\xe9phant")
-    False
-    >>> _check_unicode_camelcase(u"\xc9l\xe9Phant")
-    True
-    >>> _check_unicode_camelcase(u"\xe9l\xe9Phant")
-    False
-    >>> _check_unicode_camelcase(u"\xc9l\xe9PhanT")
-    False
-    """
-    if not pagename[0].isupper():
-        return False
-    pagename = pagename.split('@', 1)[0].split('#', 1)[0]
-    if not pagename[-1].islower():
-        return False
-    humps = 0
-    for i in xrange(1, len(pagename)):
-        if pagename[i-1].isupper():
-            if pagename[i].islower():
-                humps += 1
-            else:
-                return False
-    return humps > 1
+        >>> resource_exists(env, Resource('wiki', 'WikiStart'))
+        False
 
+        >>> from trac.wiki.model import WikiPage
+        >>> main = WikiPage(env, 'WikiStart')
+        >>> main.text = 'some content'
+        >>> main.save('author', 'no comment', '::1')
+        >>> resource_exists(env, main.resource)
+        True
+        """
+        if resource.version is None:
+            return resource.id in self.pages
+        db = self.env.get_read_db()
+        cursor = db.cursor()
+        cursor.execute("SELECT name FROM wiki WHERE name=%s AND version=%s",
+                       (resource.id, resource.version))
+        return bool(cursor.fetchall())
