@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2005-2006 Edgewall Software
+# Copyright (C) 2005-2009 Edgewall Software
 # Copyright (C) 2005-2006 Christian Boos <cboos@neuf.fr>
 # All rights reserved.
 #
@@ -16,43 +16,36 @@
 
 import re
 
+from genshi.builder import tag
+
+from trac.cache import cached
 from trac.core import *
-from trac.wiki.formatter import Formatter
+from trac.util.translation import _
 from trac.wiki.api import IWikiChangeListener, IWikiMacroProvider
+from trac.wiki.parser import WikiParser
+from trac.wiki.formatter import split_url_into_path_query_fragment
 
 
 class InterWikiMap(Component):
-    """Implements support for InterWiki maps."""
+    """InterWiki map manager."""
 
     implements(IWikiChangeListener, IWikiMacroProvider)
 
     _page_name = 'InterMapTxt'
     _interwiki_re = re.compile(r"(%s)[ \t]+([^ \t]+)(?:[ \t]+#(.*))?" %
-                               Formatter.LINK_SCHEME, re.UNICODE)
+                               WikiParser.LINK_SCHEME, re.UNICODE)
     _argspec_re = re.compile(r"\$\d")
 
-    def __init__(self):
-        self._interwiki_map = None
-        # This dictionary maps upper-cased namespaces
-        # to (namespace, prefix, title) values;
-
-    # The component itself behaves as a map
+    # The component itself behaves as a read-only map
 
     def __contains__(self, ns):
-        self._update()
-        return ns.upper() in self._interwiki_map
+        return ns.upper() in self.interwiki_map
 
     def __getitem__(self, ns):
-        self._update()
-        return self._interwiki_map[ns.upper()]
-
-    def __setitem__(self, ns, value):
-        self._update()
-        self._interwiki_map[ns.upper()] = value
+        return self.interwiki_map[ns.upper()]
 
     def keys(self):
-        self._update()
-        return self._interwiki_map.keys()
+        return self.interwiki_map.keys()
 
     # Expansion of positional arguments ($1, $2, ...) in URL and title
     def _expand(self, txt, args):
@@ -77,39 +70,51 @@ class InterWikiMap(Component):
         ns, url, title = self[ns]
         maxargnum = max([0]+[int(a[1:]) for a in
                              re.findall(InterWikiMap._argspec_re, url)])
+        target, query, fragment = split_url_into_path_query_fragment(target)
         if maxargnum > 0:
             args = target.split(':', (maxargnum - 1))
         else:
             args = [target]
-        expanded_url = self._expand_or_append(url, args)
+        url = self._expand_or_append(url, args)
+        ntarget, nquery, nfragment = split_url_into_path_query_fragment(url)
+        if query and nquery:
+            nquery = '%s&%s' % (nquery, query[1:]) 
+        else:
+            nquery = nquery or query
+        nfragment = fragment or nfragment # user provided takes precedence
+        expanded_url = ntarget + nquery + nfragment
         expanded_title = self._expand(title, args)
         if expanded_title == title:
-            expanded_title = target+' in '+title
+            expanded_title = _("%(target)s in %(name)s",
+                               target=target, name=title)
         return expanded_url, expanded_title
 
     # IWikiChangeListener methods
 
     def wiki_page_added(self, page):
-        pass
+        if page.name == InterWikiMap._page_name:
+            del self.interwiki_map
 
     def wiki_page_changed(self, page, version, t, comment, author, ipnr):
         if page.name == InterWikiMap._page_name:
-            self._interwiki_map = None
+            del self.interwiki_map
 
     def wiki_page_deleted(self, page):
         if page.name == InterWikiMap._page_name:
-            self._interwiki_map = None
+            del self.interwiki_map
 
     def wiki_page_version_deleted(self, page):
         if page.name == InterWikiMap._page_name:
-            self._interwiki_map = None
+            del self.interwiki_map
 
-    def _update(self):
+    @cached
+    def interwiki_map(self, db):
+        """Map from upper-cased namespaces to (namespace, prefix, title) 
+        values.
+        """
         from trac.wiki.model import WikiPage
-        if self._interwiki_map is not None:
-            return
-        self._interwiki_map = {}
-        content = WikiPage(self.env, InterWikiMap._page_name).text
+        map = {}
+        content = WikiPage(self.env, InterWikiMap._page_name, db=db).text
         in_map = False
         for line in content.split('\n'):
             if in_map:
@@ -121,9 +126,10 @@ class InterWikiMap(Component):
                         prefix, url, title = m.groups()
                         url = url.strip()
                         title = title and title.strip() or prefix
-                        self[prefix] = (prefix, url, title)
+                        map[prefix.upper()] = (prefix, url, title)
             elif line.startswith('----'):
                 in_map = True
+        return map
 
     # IWikiMacroProvider methods
 
@@ -133,9 +139,7 @@ class InterWikiMap(Component):
     def get_macro_description(self, name): 
         return "Provide a description list for the known InterWiki prefixes."
 
-    def render_macro(self, req, name, content):
-        from trac.util import sorted
-        from trac.util.html import html as _
+    def expand_macro(self, formatter, name, content):
         interwikis = []
         for k in sorted(self.keys()):
             prefix, url, title = self[k]
@@ -144,8 +148,10 @@ class InterWikiMap(Component):
                 'rc_url': self._expand_or_append(url, ['RecentChanges']),
                 'description': title == prefix and url or title})
 
-        return _.TABLE(_.TR(_.TH(_.EM("Prefix")), _.TH(_.EM("Site"))),
-                       [ _.TR(_.TD(_.A(w['prefix'], href=w['rc_url'])),
-                              _.TD(_.A(w['description'], href=w['url'])))
-                         for w in interwikis ],
-                       class_="wiki interwiki")
+        return tag.table(tag.tr(tag.th(tag.em("Prefix")),
+                                tag.th(tag.em("Site"))),
+                         [tag.tr(tag.td(tag.a(w['prefix'], href=w['rc_url'])),
+                                 tag.td(tag.a(w['description'],
+                                              href=w['url'])))
+                          for w in interwikis ],
+                         class_="wiki interwiki")
